# Copyright (C) 2011 Mathias Brodala
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import Gtk

import os.path

import pylast

from xl.nls import gettext as _
from xl import common, settings
from xlgui import icons
from xlgui.preferences import widgets
from xlgui.widgets import dialogs

name = _('Last.fm Loved Tracks')
basedir = os.path.dirname(os.path.realpath(__file__))
ui = os.path.join(basedir, "lastfmlove_preferences.ui")
icons.MANAGER.add_icon_name_from_directory('lastfm', os.path.join(basedir, 'icons'))
icon = 'lastfm'


class APIKeyPreference(widgets.Preference):
    name = 'plugin/lastfmlove/api_key'


class APISecretPrefence(widgets.Preference):
    name = 'plugin/lastfmlove/api_secret'


class RequestAccessPermissionButton(widgets.Button):
    name = 'plugin/lastfmlove/request_access_permission'

    def __init__(self, preferences, widget):
        """
        Sets up the message
        """
        widgets.Button.__init__(self, preferences, widget)

        self.message = dialogs.MessageBar(
            parent=preferences.builder.get_object('preferences_box'),
            buttons=Gtk.ButtonsType.CLOSE,
        )
        self.errors = {pylast.STATUS_INVALID_API_KEY: _('The API key is invalid.')}

    @common.threaded
    def check_connection(self):
        """
        Checks API key and secret for validity
        and opens the URI for access permission
        """
        api_key = settings.get_option('plugin/lastfmlove/api_key', 'K')

        try:
            pylast.LastFMNetwork(
                api_key=api_key,
                api_secret=settings.get_option('plugin/lastfmlove/api_secret', 'S'),
                username=settings.get_option('plugin/ascrobbler/user', ''),
                password_hash=settings.get_option('plugin/ascrobbler/password', ''),
            )
        except pylast.WSError as e:
            GLib.idle_add(
                self.message.show_error,
                self.errors[int(e.get_id())],
                _('Please make sure the entered data is correct.'),
            )
        else:
            application_launched = Gtk.show_uri(
                Gdk.Screen.get_default(),
                'http://www.last.fm/api/auth?api_key={0}'.format(api_key),
                Gdk.CURRENT_TIME,
            )

            if not application_launched:
                url = 'http://www.last.fm/api/auth?api_key={0}'.format(api_key)
                GLib.idle_add(
                    self.message.show_warning,
                    _('Could not start web browser'),
                    _(
                        'Please copy the following URL and '
                        'open it with your web browser:\n'
                        '<b><a href="{url}">{url}</a></b>'
                    ).format(url=url),
                )

    def on_clicked(self, button):
        """
        Initiates the check for validity
        """
        self.check_connection()



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import time
import re
from typing import Collection

from xl.unicode import shave_marks

__all__ = ['TracksMatcher', 'search_tracks']


class SearchResultTrack:
    """
    Holds a track with search result metadata included.

    :param track: The Track object
    """

    __slots__ = ['track', 'on_tags']

    def __init__(self, track):
        self.track = track
        self.on_tags = []


class _Matcher:
    """
    Base class for match conditions
    """

    __slots__ = ['tag', 'content', 'lower']

    def __init__(self, tag, content, lower):
        self.tag = tag
        if content and not self.tag.startswith("__"):
            content = lower(content)
        self.content = content
        self.lower = lower

    def match(self, srtrack):
        vals = srtrack.track.get_tag_search(self.tag, format=False)
        if vals == '__null__':
            vals = None
        if not isinstance(vals, list):
            vals = [vals]

        for item in vals:
            if item is not None:
                item = self.lower(item)

            if self._matches(item):
                return True
        return False

    def _matches(self, value):
        raise NotImplementedError


class _ExactMatcher(_Matcher):
    """
    Condition for exact matches
    """

    def _matches(self, value):
        if self.tag.startswith("__"):
            try:
                newvalue = float(value)
                newcontent = float(self.content)
                return abs(newvalue - newcontent) < 0.0001
            except (TypeError, ValueError):
                newvalue = value
                newcontent = self.content
        else:
            newvalue = value
            newcontent = self.content
        return newvalue == newcontent


class _InMatcher(_Matcher):
    """
    Condition for inexact (ie. containing) matches
    """

    def _matches(self, value):
        if not value:
            return False
        try:
            return self.content in value
        except TypeError:
            return False


class _RegexMatcher(_Matcher):
    """
    Condition for regular expression matches
    """

    def __init__(self, tag, content, lower):
        _Matcher.__init__(self, tag, content, lower)
        self._re = re.compile(content)

    def _matches(self, value):
        if not value:
            return False
        try:
            return self._re.search(value) is not None
        except TypeError:
            return False


class _GtMatcher(_Matcher):
    """
    Condition for greater than matches.
    """

    def _matches(self, value):
        try:
            value = float(value)
            content = float(self.content)  # kinda inefficient
        except (TypeError, ValueError):
            return False
        return value > content


class _LtMatcher(_Matcher):
    """
    Condition for less than matches.
    """

    def _matches(self, value):
        try:
            if value is None:
                value = 0
            else:
                value = float(value)
            content = float(self.content)  # kinda inefficient
        except (TypeError, ValueError):
            return False
        return value < content


class _NotMetaMatcher:
    """
    Condition for boolean NOT
    """

    __slots__ = ['matcher']
    tag = None

    def __init__(self, matcher):
        self.matcher = matcher

    def match(self, srtrack):
        return not self.matcher.match(srtrack)


class _OrMetaMatcher:
    """
    Condition for boolean OR
    """

    __slots__ = ['left', 'right']
    tag = None

    def __init__(self, left, right):
        self.left, self.right = left, right

    def match(self, srtrack):
        return self.left.match(srtrack) or self.right.match(srtrack)


class _MultiMetaMatcher:
    """
    Condition for boolean AND
    """

    __slots__ = ['matchers']
    tag = None

    def __init__(self, matchers):
        self.matchers = matchers

    def match(self, srtrack):
        for ma in self.matchers:
            if not ma.match(srtrack):
                return False
        return True


class _ManyMultiMetaMatcher:
    """
    TODO: think of a proper docstring for this

    This handles the case where we want to match in an OR-like
    fashion, but also know which tags were matched. Useful for
    the collection panel expansion.
    """

    __slots__ = ['matchers', 'tags']
    tag = None

    def __init__(self, matchers):
        self.matchers = matchers
        self.tags = set()

    def match(self, srtrack):
        self.tags = set()
        matched = False
        for ma in self.matchers:
            if ma.match(srtrack):
                if ma.tag:
                    matched = True
                    self.tags.add(ma.tag)
                elif hasattr(ma, 'tags') and ma.tags:
                    matched = True
                    self.tags.update(ma.tags)
        return matched


class TracksMatcher:
    """
    Holds criteria and determines whether
    a given track matches those criteria.
    """

    __slots__ = ['matchers', 'case_sensitive', 'keyword_tags']

    def __init__(self, search_string, case_sensitive=True, keyword_tags=None):
        """
        :param search_string: a string describing the match conditions
        :param case_sensitive: whether to search in a case-sensitive
            manner.
        :param keyword_tags: a list of tags to match search keywords
            in.
        """
        self.case_sensitive = case_sensitive
        self.keyword_tags = keyword_tags or []
        search_string = shave_marks(search_string)
        tokens = self.__tokenize_query(search_string)
        tokens = self.__red(tokens)
        tokens = self.__optimize_tokens(tokens)
        self.matchers = self.__tokens_to_matchers(tokens)

    def append_matcher(self, matcher, or_match=False):
        '''Here so you can use playlist matchers. Probably needs better impl'''
        if not or_match or len(self.matchers) == 0:
            self.matchers.append(matcher)
        else:
            self.matchers[-1] = _OrMetaMatcher(self.matchers[-1], matcher)

    def prepend_matcher(self, matcher, or_match=False):
        '''Here so you can use playlist matchers. Probably needs better impl'''
        if not or_match or len(self.matchers) == 0:
            self.matchers.insert(0, matcher)
        else:
            self.matchers[0] = _OrMetaMatcher(matcher, self.matchers[0])

    def match(self, srtrack):
        """
        Determine whether a given SearchResultTrack's internal
        Track object matches this search condition.
        """
        for ma in self.matchers:
            if not ma.match(srtrack):
                break
            if ma.tag is not None:
                if ma.tag not in srtrack.on_tags:
                    srtrack.on_tags.append(ma.tag)
            elif hasattr(ma, 'tags'):
                for t in ma.tags:
                    if t not in srtrack.on_tags:
                        srtrack.on_tags.append(t)
        else:
            return True
        return False

    def __tokens_to_matchers(self, tokens, matchers=None):
        """
        Converts a token hierarchy to a list of matchers
        """
        if not matchers:
            matchers = []

        # if there's no more tokens, we're done
        try:
            token = tokens[0]
        except IndexError:
            return matchers

        # is it a special operator?
        if isinstance(token, list):
            if len(token) == 1:
                token = token[0]
            subtoken = token[0]
            # NOT
            if subtoken == "!":
                nots = self.__tokens_to_matchers(token[1])
                matchers.append(_NotMetaMatcher(_MultiMetaMatcher(nots)))
            # OR
            elif subtoken == "|":
                left = self.__tokens_to_matchers([token[1][0]])
                right = self.__tokens_to_matchers([token[1][1]])
                matchers.append(
                    _OrMetaMatcher(_MultiMetaMatcher(left), _MultiMetaMatcher(right))
                )
            # ()
            elif subtoken == "(":
                inner = self.__tokens_to_matchers([token[1]])
                matchers.append(_MultiMetaMatcher(inner))
            else:
                return matchers

        elif token == '':
            pass

        # normal token
        else:
            if not self.case_sensitive:
                lower = lambda x: x.lower()
            else:
                lower = lambda x: x

            # TODO: this stuff is kinda repetitive, can we consolidate
            # it? Maybe move some of this into the matcher classes?

            # exact match in tag
            if "==" in token:
                tag, content = token.split("==", 1)
                if content == "__null__":
                    content = None
                matcher = _ExactMatcher(tag, content, lower)
                matchers.append(matcher)

            # keyword in tag
            elif "=" in token:
                tag, content = token.split("=", 1)
                content = content.strip().strip('"')
                matcher = _InMatcher(tag, content, lower)
                matchers.append(matcher)

            elif ">" in token:
                tag, content = token.split(">", 1)
                content = content.strip().strip('"')
                matcher = _GtMatcher(tag, content, lower)
                matchers.append(matcher)

            elif "<" in token:
                tag, content = token.split("<", 1)
                content = content.strip().strip('"')
                matcher = _LtMatcher(tag, content, lower)
                matchers.append(matcher)

            elif "~" in token:
                tag, content = token.split("~", 1)
                content = content.strip().strip('"')
                matcher = _RegexMatcher(tag, content, lower)
                matchers.append(matcher)

            # plain keyword
            else:
                content = token.strip().strip('"')
                mmm = []
                for tag in self.keyword_tags:
                    matcher = _InMatcher(tag, content, lower)
                    mmm.append(matcher)
                matchers.append(_ManyMultiMetaMatcher(mmm))

        return self.__tokens_to_matchers(tokens[1:], matchers)

    def __tokenize_query(self, search):
        """
        Turns a search string into a list of tokens.
        """
        search = " " + search + " "

        tokens = []
        newsearch = ""
        in_quotes = False
        in_regex = False
        n = 0
        while n < len(search):
            c = search[n]
            if c == "\\":
                if not in_regex:
                    n += 1
                try:
                    newsearch += search[n]
                except IndexError:
                    pass
            elif in_quotes and c != "\"":
                newsearch += c
            elif c == "~":
                in_regex = True
                newsearch += c
            elif c == "\"":
                in_quotes = not in_quotes  # toggle
                # newsearch += c
            elif c in ["|", "!", "(", ")"]:
                newsearch += c
            elif c == " ":
                in_regex = False
                tokens.append(newsearch)
                newsearch = ""
            else:
                newsearch += c
            n += 1
        return tokens

    def __red(self, tokens):
        """
        Turn the token list into a token list hierarchy that is
        easier to parse.
        """
        # base case since we use recursion
        if tokens == []:
            return []

        # handle parentheses
        elif "(" in tokens:
            num_found = 0
            start = None
            end = None
            count = 0
            for t in tokens:
                if t == "(":
                    if start is None:
                        start = count
                    else:
                        num_found += 1
                elif t == ")":
                    if end is None and num_found == 0:
                        end = count
                    else:
                        num_found -= 1
                if start and end:
                    break
                count += 1
            before = tokens[:start]
            inside = self.__red(tokens[start + 1 : end])
            after = tokens[end + 1 :]
            tokens = before + [["(", inside]] + after

        # handle NOT
        elif "!" in tokens:
            start = tokens.index("!")
            end = start + 2
            before = tokens[:start]
            inside = tokens[start + 1 : end]
            after = tokens[end:]
            tokens = before + [["!", inside]] + after

        # handle OR
        elif "|" in tokens:
            start = tokens.index("|")
            inside = [tokens[start - 1], tokens[start + 1]]
            before = tokens[: start - 1]
            after = tokens[start + 2 :]
            tokens = before + [["|", inside]] + after

        # nothing special, so just return it
        else:
            return tokens

        return self.__red(tokens)

    def __optimize_tokens(self, tokens):
        """
        Attempt to optimize tokens, to speed up matching.
        """
        # longer queries tend to reject more tracks, which speeds up
        # processing, so we put them first.
        tokens.sort(key=len)
        return tokens


class TracksInList:
    """
    Matches tracks contained in a list/dict/set. Copies the list.
    """

    __slots__ = ['_tracks']
    tag = None

    def __init__(self, tracks):
        if isinstance(tracks, dict):
            self._tracks = set(tracks.keys())
        else:
            self._tracks = {t for t in tracks}

    def match(self, track):
        return track.track in self._tracks


class TracksNotInList(TracksInList):
    """
    Matches tracks not in a list/dict/set
    """

    def match(self, track):
        return track.track not in self._tracks


def search_tracks(trackiter, trackmatchers: Collection[TracksMatcher]):
    """
    Search a set of tracks for those that match specified conditions.

    :param trackiter: An iterable object returning Track objects
    :param trackmatchers: A list of TrackMatcher objects
    """
    for srtr in trackiter:
        if not isinstance(srtr, SearchResultTrack):
            srtr = SearchResultTrack(srtr)
        if all(tma.match(srtr) for tma in trackmatchers):
            yield srtr

        # On large collections, searching can take a while. Due to
        # peculiarities in python's GIL that means the now-cpu-bound
        # thread running the search can end up blocking other threads.
        # Calling out to time.sleep forces a release of the GIL and
        # allows other threads to run. Benchmarks show this has no
        # noticable effect on search speed.
        time.sleep(0)


def search_tracks_from_string(
    trackiter, search_string, case_sensitive=True, keyword_tags=None
):
    """
    Convenience wrapper around search_tracks that builds matchers
    automatically from the search string.

    Arguments have the same meaning as the corresponding arguments on
    on :class:`search_tracks` and :class:`TracksMatcher`.
    """
    matchers = [
        TracksMatcher(
            search_string, case_sensitive=case_sensitive, keyword_tags=keyword_tags
        )
    ]
    return search_tracks(trackiter, matchers)


def match_track_from_string(
    track, search_string, case_sensitive=True, keyword_tags=None
):
    matcher = TracksMatcher(
        search_string, case_sensitive=case_sensitive, keyword_tags=keyword_tags
    )
    return matcher.match(SearchResultTrack(track))



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import logging
import threading
import time
import dbus

from xl import providers, event
from xl.main import common

logger = logging.getLogger(__name__)


class UDisksPropertyWrapper:
    """
    Wrapper around an org.freedesktop.DBus.Properties interface

    You shouldn't need to create this, use UDisksDBusWrapper.props
    """

    def __init__(self, obj, iface_type):
        self.obj = obj  # properties object
        self.iface_type = iface_type

    def __getattr__(self, name):
        return lambda *a, **k: self.obj.__getattr__(name)(
            *((self.iface_type,) + a), **k
        )

    # def connect_on_changed(self, fn):
    #    '''Connect to the PropertiesChanged signal'''
    #    return self.obj.connect_to_signal('PropertiesChanged', fn, self.iface_type)

    def __repr__(self):
        return '<UDisksPropertyWrapper: %s>' % self.iface_type


class UDisksDBusWrapper:
    """
    Simple wrapper to make life easier. Assume that we only are
    using this to get properties off the 'primary' interface.

    Assumes that you are only asking for properties on the interface
    associated with the object path.

    Example usage:

        obj = get_object_by_path('/org/freedesktop/UDisks2/drives/foo')
        print(obj.props.Get('Device'))

    You shouldn't need to create this, use get_object_py_path.
    """

    __slots__ = ['obj', 'iface_type', '_iface', '_props_iface', 'path']

    def __init__(self, bus, root, path, iface_type):
        self.obj = bus.get_object(root, path)
        self.iface_type = iface_type
        self._iface = None
        self._props_iface = None

    def __getattr__(self, member):
        return self.iface.__getattr__(member)

    def connect_to_signal(self, *a, **k):
        '''Connect to a signal on the object's primary interface type'''
        return self.iface.connect_to_signal(*a, **k)

    @property
    def iface(self):
        '''Returns a dbus.Interface for the object's primary interface type'''
        if self._iface is None:
            self._iface = dbus.Interface(self.obj, self.iface_type)
        return self._iface

    @property
    def object_path(self):
        return self.obj.object_path

    @property
    def props(self):
        '''Returns a dbus.Interface for org.freedesktop.DBus.Properties'''
        if self._props_iface is None:
            iface = dbus.Interface(self.obj, 'org.freedesktop.DBus.Properties')
            self._props_iface = UDisksPropertyWrapper(iface, self.iface_type)
        return self._props_iface

    def __repr__(self):
        return '<UDisksDBusWrapper: %s (%s)>' % (self.iface_type, self.path)


class UDisksBase(providers.ProviderHandler):
    """
    Provides support for UDisks (1 and 2) devices. To get properties
    of devices, the get_object_for_path function will return a convenient
    wrapper to use for accessing properties and such on it.

    Implements the udisks and udisks2 service for providers. Providers
    should override the UDisksProvider interface.

    Plugins should not try to connect to UDisks until exaile has finished
    loading.
    """

    # States: start -> init -> addremove <-> listening -> end.
    # The addremove state acts as a lock against concurrent changes.

    def __init__(self, devicemanager):
        self._lock = threading.Lock()
        self._state = 'init'

        providers.ProviderHandler.__init__(self, self.name)
        self.devicemanager = devicemanager

        self.bus = None
        self.devices = {}
        self.providers = {}

    def connect(self):
        assert self._state == 'init'
        logger.debug("Connecting to %s", self.name)
        try:
            self.obj = self._connect()
            logger.info("Connected to %s", self.name)
            event.log_event("hal_connected", self, None)
        except Exception:
            logger.info(
                "Failed to connect to %s, automatic detection of "
                "devices will be disabled.",
                self.name,
            )
            return False

        self._state = 'addremove'
        logger.debug("%s: state = addremove", self.name)
        self._add_all(self.obj)
        self._state = 'listening'
        logger.debug("%s: state = listening", self.name)
        return True

    #
    # Public API
    #

    def get_object_by_path(self, path):
        """
        Call this to retrieve a UDisksDBusWrapper object for the path
        of the object you want to retrieve.

        :param path: The udisks path of the object you want to retrieve

        :returns: UDisksDBusWrapper object
        :raises: KeyError if the object path/type is not supported
        """

        for p, iface_type in self.paths:
            if path.startswith(p):
                return UDisksDBusWrapper(self.bus, self.root, path, iface_type)

        raise KeyError("Unsupported path %s" % path)

    #
    # subclasses must implement these
    #

    def _connect(self):
        raise NotImplementedError()

    def _add_all(self, obj):
        raise NotImplementedError()

    #
    # Private API
    #

    def _add_device(self, path, obj=None):
        """
        Call with either path or obj (obj gets priority). Not thread-safe.
        """
        assert self._state == 'addremove'

        if obj is None:
            obj = self.get_object_by_path(path)

        # In the following code, `old` and `new` are providers, while
        # `self.devices[path]` and `device` are old/new devices. There are
        # several possible code paths that should be correctly handled:
        # - No old nor new provider for this path.
        # - Provider changes (nothing to something, something to nothing,
        #   something to something else); obviously device changes as well.
        # - Provider stays the same, but device changes (i.e. instant media-
        #   swapping; not sure it can happen).
        # - Provider and device stay the same.
        old, new = self._get_provider_for(obj)

        if new is None:
            if old is not None:
                self._remove_device(path)
            return

        device = new.get_device(obj, self)
        if new is old and device is self.devices[path]:
            return  # Exactly the same device

        if old is not None:
            self._remove_device(path)

        if new is None:
            return
        try:
            device.autoconnect()
        except Exception:
            logger.exception("%s: Failed autoconnecting device %s", self.name, device)
        else:
            self.devicemanager.add_device(device)
            self.providers[path] = new
            self.devices[path] = device

    def _on_change(self, path):

        assert self._state == 'addremove'

        obj = self.get_object_by_path(path)

        provider = self.providers.get(obj.object_path)
        if provider is None:
            self._add_device(path, obj)
        else:
            remove = provider.on_device_changed(obj, self, self.devices[path])
            if remove == 'remove':
                self._remove_device(path)

    def _get_provider_for(self, obj):
        """
        Return (old_provider, old_priority), (new_provider, new_priority).
        Not thread-safe.
        """
        assert self._state == 'addremove'
        highest_prio = -1
        highest = None
        old = self.providers.get(obj.object_path)
        for provider in self.get_providers():
            priority = provider.get_priority(obj, self)
            if priority is None:
                continue
            # Find highest priority, preferring old provider.
            if priority > highest_prio or (
                priority == highest_prio and provider is old
            ):
                highest_prio = priority
                highest = provider
        return old, highest

    def _remove_device(self, path):
        assert self._state == 'addremove'

        self.devicemanager.remove_device(self.devices[path])
        del self.devices[path]
        del self.providers[path]

    def _udisks_device_added(self, *args):
        path = args[0]
        logger.debug("%s: Device added: %s", self.name, path)
        if self._addremove():
            try:
                self._add_device(path)
            finally:
                self._state = 'listening'
                logger.debug("%s: state = listening (_device_added)", self.name)

    def _udisks_device_changed(self, *args):
        path = args[0]
        logger.debug("%s: Device changed: %s", self.name, path)
        if self._addremove():
            try:
                self._on_change(path)
            finally:
                self._state = 'listening'
                logger.debug("%s: state = listening (_device_added)", self.name)

    def _udisks_device_removed(self, *args):
        path = args[0]
        if self._addremove():
            try:
                self._remove_device(path)
                logger.debug("%s: Device removed: %s", path, self.name)
            except KeyError:  # Not ours
                pass
            finally:
                self._state = 'listening'
                logger.debug("%s: state = listening (_device_removed)", self.name)

    def on_provider_added(self, provider):
        if self._addremove():
            try:
                self._add_all(self.obj)
            finally:
                self._state = 'listening'
                logger.debug("%s: state = listening (_provider_added)", self.name)

    def on_provider_removed(self, provider):
        if self._addremove():
            try:
                to_remove = []
                for path, provider_ in self.providers.items():
                    if provider_ is provider:
                        to_remove.append(path)

                for path in to_remove:
                    self._remove_device(path)
            finally:
                self._state = 'listening'
                logger.debug("%s: state = listening (_provider_removed)", self.name)

    def _addremove(self):
        """
        Helper to transition safely from listening to addremove state.

        Returns whether the transition happens.
        """
        i = 0
        while True:
            with self._lock:
                if self._state == 'listening':
                    self._state = 'addremove'
                    logger.debug("%s: state = addremove", self.name)
                    return True

            if i == 50:
                logger.error(
                    "%s: Failed to acquire lock. Ignoring device event.", self.name
                )
                return False
            i += 1
            time.sleep(0.1)


class UDisks2(UDisksBase):

    name = 'udisks2'
    root = 'org.freedesktop.UDisks2'
    paths = [
        ('/org/freedesktop/UDisks2/block_devices/', 'org.freedesktop.UDisks2.Block'),
        ('/org/freedesktop/UDisks2/drives/', 'org.freedesktop.UDisks2.Drive'),
        ('/org/freedesktop/UDisks2', 'org.freedesktop.DBus.ObjectManager'),
    ]

    def _connect(self):

        self.bus = dbus.SystemBus()
        obj = self.get_object_by_path('/org/freedesktop/UDisks2')

        obj.connect_to_signal('InterfacesAdded', self._udisks_device_added)
        obj.connect_to_signal('InterfacesRemoved', self._udisks_device_removed)

        # listen for PropertiesChanged events on any UDisks2 object
        self.bus.add_signal_receiver(
            self._udisks2_properties_changed,
            signal_name='PropertiesChanged',
            dbus_interface='org.freedesktop.DBus.Properties',
            bus_name='org.freedesktop.UDisks2',
            path_keyword='path',
        )

        return obj

    def _udisks2_properties_changed(self, *args, **kwargs):
        # TODO: this is inefficient, probably should just let the
        #       provider know what properties changed. would need
        #       to have the consumer let us know what to subscribe to
        self._udisks_device_changed(kwargs['path'])

    def _add_all(self, obj):
        assert self._state == 'addremove'
        for path in obj.GetManagedObjects():
            self._add_device(path)


class Handler:
    """
    The HAL provider interface
    """

    name = 'base'

    def __init__(self):
        pass

    def is_type(self, device, capabilities):
        return False

    def get_udis(self, hal):
        return []

    def device_from_udi(self, hal, udi):
        pass


class UDisksProvider:
    """
    The UDisksProvider interface. Works for UDisks 1 and 2, but you should
    implement separate providers for each, as the object types and
    properties are different.

    This API is subject to change.
    """

    VERY_LOW, LOW, NORMAL, HIGH, VERY_HIGH = range(0, 101, 25)

    def get_priority(self, obj, udisks):
        """
        Called on initial connect of a device. The provider should
        return a priority value indicating its interest in handling
        the device.

        :param obj: A UDisksPropertyWrapper object for the device path
        :param udisks: The UDisksBase object

        :returns: An integer [0..100] indicating priority, or None if it
                  cannot handle the device
        """

    def get_device(self, obj, udisks):
        """
        Called when the device is assigned to the provider (e.g., it
        indicated the highest priority).

        :param obj: A UDisksPropertyWrapper object for the device path
        :param udisks: The UDisksBase object

        :returns: xl.devices.Device derived object for the device
        """

    def on_device_changed(self, obj, udisks, device):
        """
        Called when UDisks indicates that a property of the device has
        changed. If useful, the provider should forward relevant change
        actions to its device object.

        :param obj: A UDisksPropertyWrapper object for the device path
        :param udisks: The UDisksBase object
        :param device: Object returned from get_device

        :returns: 'remove' to remove the device from the provider, other
                  values ignored.
        """


# vim: et sts=4 sw=4



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import Gtk

from xl import main, playlist, settings

from xl.nls import gettext as _

from . import dialogs
from .filter import (
    EntryField,
    FilterDialog,
    ComboEntryField,
    MultiEntryField,
    NullField,
    QuotedEntryField,
    SpinButtonAndComboField,
    SpinLabelField,
)

import logging

logger = logging.getLogger(__name__)


def N_(x):
    return x


class EntrySecondsField(MultiEntryField):
    def __init__(self):
        MultiEntryField.__init__(self, (50, _('seconds')))


class EntryAndEntryField(MultiEntryField):
    def __init__(self):
        # TRANSLATORS: Logical AND used for smart playlists
        MultiEntryField.__init__(self, (50, _('and'), 50))


class EntryDaysField(MultiEntryField):
    def __init__(self):
        MultiEntryField.__init__(self, (50, _('days')))


class PlaylistField(ComboEntryField):
    def __init__(self):
        playlists = []
        playlists.extend(main.exaile().smart_playlists.list_playlists())
        playlists.extend(main.exaile().playlists.list_playlists())
        playlists.sort()
        ComboEntryField.__init__(self, playlists)


DATE_FIELDS = [N_('seconds'), N_('minutes'), N_('hours'), N_('days'), N_('weeks')]


class SpinDateField(SpinButtonAndComboField):
    def __init__(self):
        SpinButtonAndComboField.__init__(self, DATE_FIELDS)


class SpinSecondsField(SpinLabelField):
    def __init__(self):
        SpinLabelField.__init__(self, _('seconds'))


class SpinRating(SpinLabelField):
    def __init__(self):
        SpinLabelField.__init__(self, '', settings.get_option("rating/maximum", 5), 0)


class SpinNothing(SpinLabelField):
    def __init__(self):
        SpinLabelField.__init__(self, '')


# This sets up the CRITERIA for all the available types of tags
# that exaile supports. The actual CRITERIA dict is populated
# using xl.metadata.tags.tag_data.
#
# NOTE: The following strings are already marked for translation in _TRANS and
# _NMAP, and will be really translated by filtergui; no need to clutter the
# code here.
_criteria_types = {
    # fmt: off

    # TODO
    'bitrate': [
        ('is', SpinNothing),
        ('less than', SpinNothing),
        ('greater than', SpinNothing),
        ('between', EntryAndEntryField),
        ('at least', SpinNothing),
        ('at most', SpinNothing),
        ('is set', NullField),
        ('is not set', NullField),
    ],

    'image': None,

    'int': [
        ('is', SpinNothing),
        ('less than', SpinNothing),
        ('greater than', SpinNothing),
        ('between', EntryAndEntryField),
        ('at least', SpinNothing),
        ('at most', SpinNothing),
        ('is set', NullField),
        ('is not set', NullField),
    ],

    'location': [
        ('is', QuotedEntryField),
        ('is not', QuotedEntryField),
        ('contains', QuotedEntryField),
        ('does not contain', QuotedEntryField),
        ('regex', QuotedEntryField),
        ('not regex', QuotedEntryField),
    ],

    'text': [
        ('is', EntryField),
        ('is not', EntryField),
        ('contains', EntryField),
        ('does not contain', EntryField),
        ('contains word', EntryField),
        ('does not contain word', EntryField),
        ('regex', EntryField),
        ('not regex', EntryField),
        ('is set', NullField),
        ('is not set', NullField),
    ],

    'time': [
        ('at least', SpinSecondsField),
        ('at most', SpinSecondsField),
        ('is', SpinSecondsField),
        ('is not', SpinSecondsField),
    ],

    'timestamp': [
        ('in the last', SpinDateField),
        ('not in the last', SpinDateField),
    ],

    '__playlist': [
        ('Track is in', PlaylistField),
        ('Track not in', PlaylistField),
    ],

    '__rating': [
        ('greater than', SpinRating),
        ('less than', SpinRating),
        ('at least', SpinRating),
        ('at most', SpinRating),
    ],
    # fmt: on
}

# aliases
_criteria_types['datetime'] = _criteria_types['text']  # TODO: fix
_criteria_types['multiline'] = _criteria_types['text']
_criteria_types['dblnum'] = _criteria_types['int']


# This gets populated below. Only add special tags/searches that don't have a
# valid entry in tag_data
CRITERIA = [
    ('Rating', _criteria_types['__rating']),
    ('Playlist', _criteria_types['__playlist']),
]

# NOTE: We use N_ (fake gettext) because these strings are translated later by
# the filter GUI. If we use _ (real gettext) here, filtergui will try to
# translate already-translated strings, which makes no sense. This is partly due
# to the old design of storing untranslated strings (instead of operators) in
# the dynamic playlist database.

_TRANS = {
    # TRANSLATORS: True if haystack is equal to needle
    N_('is'): '==',
    # TRANSLATORS: True if haystack is not equal to needle
    N_('is not'): '!==',
    # TRANSLATORS: True if the specified tag is present (uses the NullField
    # to compare to __null__)
    N_('is set'): '<!==>',
    # TRANSLATORS: True if the specified tag is not present (uses the NullField
    # to compare to __null__)
    N_('is not set'): '<==>',
    # TRANSLATORS: True if haystack contains needle
    N_('contains'): '=',
    # TRANSLATORS: True if haystack does not contain needle
    N_('does not contain'): '!=',
    # TRANSLATORS: True if haystack contains whole word
    N_('contains word'): 'w=',
    # TRANSLATORS: True if haystack does not contain whole word
    N_('does not contain word'): '!w=',
    # TRANSLATORS: True if haystack matches regular expression
    N_('regex'): '~',
    # TRANSLATORS: True if haystack does not match regular expression
    N_('not regex'): '!~',
    # TRANSLATORS: Example: rating >= 5
    N_('at least'): '>=',
    # TRANSLATORS: Example: rating <= 3
    N_('at most'): '<=',
    # TRANSLATORS: Example: year < 1999
    N_('before'): '<',
    # TRANSLATORS: Example: year > 2002
    N_('after'): '>',
    # TRANSLATORS: Example: 1980 <= year <= 1987
    N_('between'): '><',
    N_('greater than'): '>',
    N_('less than'): '<',
    # TRANSLATORS: Example: track has been added in the last 2 days
    N_('in the last'): '>=',
    # TRANSLATORS: Example: track has not been added in the last 5 hours
    N_('not in the last'): '<',
    # TRANSLATORS: True if a track is contained in the specified playlist
    N_('Track is in'): 'pin',
    # TRANSLATORS: True if a track is not contained in the specified playlist
    N_('Track not in'): '!pin',
}

# This table is a reverse lookup for the actual tag name from a display
# name.
# This gets populated below. Only add special tags/searches here.
_NMAP = {
    N_('Rating'): '__rating',  # special
    N_('Playlist'): '__playlist',  # not a real tag
}

_REV_NMAP = {}


# update the tables based on the globally stored tag list
def __update_maps():

    from xl.metadata.tags import tag_data

    for tag, data in tag_data.items():

        if data is None:
            continue

        # don't catch this KeyError -- if it fails, fix it!
        criteria = _criteria_types[data.type]

        if criteria is None:
            continue

        CRITERIA.append((data.name, criteria))

        _NMAP[data.name] = tag

    for k, v in _NMAP.items():
        if v in _REV_NMAP:
            raise ValueError("_REV_NMAP Internal error: '%s', '%s'" % (k, v))
        _REV_NMAP[v] = k


__update_maps()


class SmartPlaylistEditor:
    @classmethod
    def create(cls, collection, smart_manager, parent=None):
        """
        Shows a dialog to create a new smart playlist

        :param collection:    Collection object
        :param smart_manager: SmartPlaylistManager object
        :param parent:        Dialog parent

        :returns: New smart playlist, or None
        """
        dialog = FilterDialog(_('Add Smart Playlist'), parent, CRITERIA)
        dialog.set_transient_for(parent)

        cls._attach_sort_widgets(dialog, None, None)

        pl = cls._run_edit_dialog(dialog, collection, smart_manager, parent)
        dialog.destroy()
        return pl

    @classmethod
    def edit(cls, pl, collection, smart_manager, parent=None):
        """
        Shows a dialog to edit a smart playlist

        :param collection:    Collection object
        :param smart_manager: SmartPlaylistManager object
        :param parent:        Dialog parent

        :returns: New smart playlist, or None
        """
        if not isinstance(pl, playlist.SmartPlaylist):
            return

        from xl.metadata.tags import tag_data

        params = pl.search_params
        state = []

        for param in params:
            (field, op, value) = param
            rev_field = _REV_NMAP[field]

            # because there are duplicates in _TRANS, cannot create a reverse
            # mapping. Instead, search in set of criteria defined for the type
            data_type = field
            data = tag_data.get(field)
            if data is not None:
                data_type = data.type

            for ct in _criteria_types[data_type]:
                rev_op = ct[0]
                if _TRANS[rev_op] == op:
                    break
            else:
                dialogs.error(parent, "Invalid operand for %s, omitting" % rev_field)
                continue

            state.append(([rev_field, rev_op], value))

        state.reverse()

        dialog = FilterDialog(_('Edit Smart Playlist'), parent, CRITERIA)

        dialog.set_transient_for(parent)
        dialog.set_name(pl.name)
        dialog.set_match_any(pl.get_or_match())
        dialog.set_limit(pl.get_return_limit())
        dialog.set_random(pl.get_random_sort())
        dialog.set_state(state)

        cls._attach_sort_widgets(dialog, *pl.get_sort_tags())

        pl = cls._run_edit_dialog(dialog, collection, smart_manager, parent, orig_pl=pl)
        dialog.destroy()
        return pl

    @classmethod
    def _run_edit_dialog(cls, dialog, collection, smart_manager, parent, orig_pl=None):
        '''internal helper function'''

        while True:
            result = dialog.run()
            dialog.hide()

            if result != Gtk.ResponseType.ACCEPT:
                return

            name = dialog.get_name()
            matchany = dialog.get_match_any()
            limit = dialog.get_limit()
            state = dialog.get_state()
            random = dialog.get_random()
            sort_tags = cls._get_sort_tags(dialog)

            if not name:
                dialogs.error(
                    parent, _("You did " "not enter a name for your playlist")
                )
                continue

            if not orig_pl or name != orig_pl.name:
                try:
                    pl = smart_manager.get_playlist(name)
                    dialogs.error(
                        parent, _("The " "playlist name you entered is already in use.")
                    )
                    continue
                except ValueError:
                    pass  # playlist didn't exist

            pl = playlist.SmartPlaylist(name, collection)
            pl.set_or_match(matchany)
            pl.set_return_limit(limit)
            pl.set_random_sort(random)
            pl.set_sort_tags(*sort_tags)

            for item in state:
                (field, op) = item[0]
                value = item[1]
                pl.add_param(_NMAP[field], _TRANS[op], value)

            if orig_pl:
                smart_manager.remove_playlist(pl.name)

            smart_manager.save_playlist(pl)
            return pl

    @classmethod
    def _attach_sort_widgets(cls, dialog, tag, reverse):
        # Add sort widgets
        from xl.metadata.tags import tag_data

        dialog.sort_enable = sort_enable = Gtk.CheckButton.new_with_label(_('Sort by:'))

        def _on_sort_enable_changed(ck):
            sort_tags.set_sensitive(ck.get_active())
            sort_order.set_sensitive(ck.get_active())

        sort_enable.connect('toggled', _on_sort_enable_changed)

        dialog.sort_tags = sort_tags = Gtk.ComboBoxText.new()
        sort_tags.set_wrap_width(5)
        for k, v in sorted(tag_data.items()):
            if v:
                sort_tags.append(k, v.translated_name)

        dialog.sort_order = sort_order = Gtk.ComboBoxText.new()
        sort_order.append("False", _("Ascending"))
        sort_order.append("True", _("Descending"))

        box = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 5)
        box.pack_start(sort_enable, False, False, 0)
        box.pack_start(sort_tags, False, False, 0)
        box.pack_start(sort_order, False, False, 0)
        box.show_all()

        dialog.get_content_area().pack_start(box, False, False, 5)

        # Set it up
        if tag is None:
            sort_enable.set_active(False)
            sort_order.set_active_id("False")
        else:
            sort_enable.set_active(True)
            sort_tags.set_active_id(tag)
            sort_order.set_active_id(str(reverse))

        _on_sort_enable_changed(sort_enable)

    @classmethod
    def _get_sort_tags(cls, dialog):
        if not dialog.sort_enable.get_active():
            return None, False
        tag = dialog.sort_tags.get_active_id()
        order = dialog.sort_order.get_active_id() == 'True'
        return tag, order



# encoding:utf-8

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import logging
import os
import urllib.error
import urllib.parse

from gi.repository import Gtk
from gi.repository import GLib

from xl import common, event, providers, settings
from xl.nls import gettext as _
from xlgui import panel

from . import preferences

import gi

gi.require_version('WebKit2', '4.0')
from gi.repository import WebKit2


log = logging.getLogger(__name__)

# fmt: off
LANGUAGES = ["ab", "aa", "af", "ak", "sq", "am", "ar", "an", "hy", "as", "av",
             "ae", "ay", "az", "bm", "ba", "eu", "be", "bn", "bh", "bi", "bs", "br", "bg",
             "my", "ca", "ch", "ce", "ny", "cv", "kw", "co", "cr", "hr", "cs", "da", "dv",
             "nl", "dz", "en", "eo", "et", "ee", "fo", "fj", "fi", "fr", "ff", "gl", "ka",
             "de", "el", "gn", "gu", "ht", "ha", "he", "hz", "hi", "ho", "hu", "ia", "id",
             "ie", "ga", "ig", "ik", "io", "is", "it", "iu", "jv", "kl", "kn", "kr", "kk",
             "km", "ki", "rw", "ky", "kv", "kg", "kj", "la", "lb", "lg", "li", "ln", "lo",
             "lt", "lv", "gv", "mk", "mg", "ml", "mt", "mi", "mr", "mh", "mn", "na", "nv",
             "nb", "nd", "ne", "ng", "nn", "no", "ii", "nr", "oc", "oj", "cu", "om", "or",
             "os", "pi", "fa", "pl", "ps", "pt", "qu", "rm", "rn", "ro", "ru", "sa", "sc",
             "se", "sm", "sg", "sr", "gd", "sn", "si", "sk", "sl", "so", "st", "es", "su",
             "sw", "ss", "sv", "ta", "te", "th", "ti", "bo", "tk", "tl", "tn", "to", "tr",
             "ts", "tw", "ty", "uk", "ur", "ve", "vi", "vk", "vo", "wa", "cy", "wo", "fy",
             "xh", "yi", "yo", "za", "zu"]
# fmt: on


class WikipediaPlugin:

    __exaile = None
    __wiki_panel = None

    def enable(self, exaile):
        self.__exaile = exaile

    def disable(self, _exaile):
        providers.unregister('main-panel', self.__wiki_panel)
        self.__wiki_panel.destroy()
        self.__exaile = None
        self.__wiki_panel = None

    def on_gui_loaded(self):
        user_agent = self.__exaile.get_user_agent_string('wikipedia')
        self.__wiki_panel = WikiPanel(self.__exaile.gui.main.window, user_agent)
        providers.register('main-panel', self.__wiki_panel)

    def get_preferences_pane(self):
        return preferences


plugin_class = WikipediaPlugin


class BrowserPage(WebKit2.WebView):
    def __init__(self, builder, user_agent):
        WebKit2.WebView.__init__(self)

        self.hometrack = None
        self.__user_agent = user_agent

        builder.connect_signals(self)
        event.add_callback(self.on_playback_start, 'playback_track_start')

    def destroy(self):
        event.remove_callback(self.on_playback_start, 'playback_track_start')

    def on_playback_start(self, type, player, track):
        self.hometrack = track
        self.load_wikipedia_page(track)

    def on_home_button_clicked(self, button):
        if self.hometrack is not None:
            self.load_wikipedia_page(self.hometrack)

    def on_refresh_button_clicked(self, button):
        self.reload()

    def on_back_button_clicked(self, button):
        self.go_back()

    def on_forward_button_clicked(self, button):
        self.go_forward()

    @common.threaded
    def load_wikipedia_page(self, track):
        if track != self.hometrack:
            return

        artist = track.get_tag_display('artist')
        language = settings.get_option('plugin/wikipedia/language', 'en')
        if language not in LANGUAGES:
            log.error('Provided language "%s" not found.' % language)
            language = 'en'

        artist = urllib.parse.quote(artist.encode('utf-8'), '')
        url = "https://%s.m.wikipedia.org/wiki/Special:Search/%s" % (language, artist)

        try:
            html = common.get_url_contents(url, self.__user_agent)
        except urllib.error.URLError as e:
            log.error(e)
            log.error(
                "Error occurred when trying to retrieve Wikipedia page "
                "for %s." % artist
            )
            html = (
                """
                <p style="color: red">No Wikipedia page found for <strong>%s</strong></p>
                """
                % artist
            )

        GLib.idle_add(self.load_html, html, url)


class WikiPanel(panel.Panel):
    # Specifies the path to the UI file and the name of the root element
    ui_info = (os.path.dirname(__file__) + "/data/wikipanel.ui", 'WikiPanel')

    def __init__(self, parent, user_agent):
        panel.Panel.__init__(self, parent, 'wikipedia', _('Wikipedia'))
        self.parent = parent
        self._browser = BrowserPage(self.builder, user_agent)
        self.setup_widgets()

    def destroy(self):
        self._browser.destroy()

    def setup_widgets(self):
        self._scrolled_window = Gtk.ScrolledWindow()
        self._scrolled_window.add(self._browser)
        frame = self.builder.get_object('rendering_frame')
        self._scrolled_window.show_all()
        frame.add(self._scrolled_window)



# Copyright (C) 2009-2010 Erin Drummond
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import json
from . import jamtree
import threading

from xl import common

USER_AGENT = None


def set_user_agent(s):
    global USER_AGENT
    USER_AGENT = s


def get_json(url):
    return json.loads(common.get_url_contents(url, USER_AGENT))


# Gets a list of jamtree.Artist objects matching the specified criteria
class get_artist_list(threading.Thread):
    def __init__(self, search_term, order_by, num_results, callback):
        threading.Thread.__init__(self)
        self.search_term = search_term
        self.order_by = order_by
        self.num_results = num_results
        self.callback = callback

    def run(self):
        url = (
            "http://api.jamendo.com/get2/name+id/artist/json/?searchquery=%s&order=%s&n=%s"
            % (self.search_term, self.order_by, self.num_results)
        )
        # print('get_artist_list: %s' % url)
        results = get_json(url)
        artists = []
        for result in results:
            item = jamtree.Artist(result['id'], result['name'].strip())
            artists.append(item)
        if artists == []:
            artists = None

        self.callback(artists)


# Gets a list of jamtree.Album objects matching the specified criteria


class get_album_list(threading.Thread):
    def __init__(self, search_term, order_by, num_results, callback):
        threading.Thread.__init__(self)
        self.search_term = search_term
        self.order_by = order_by
        self.num_results = num_results
        self.callback = callback

    def run(self):
        url = (
            "http://api.jamendo.com/get2/name+id/album/json/?searchquery=%s&order=%s&n=%s"
            % (self.search_term, self.order_by, self.num_results)
        )
        results = get_json(url)
        albums = []
        for result in results:
            ar = jamtree.Album(result['id'], result['name'].strip())
            albums.append(ar)

        if albums == []:
            albums = None

        self.callback(albums)


# Gets a list of jamtree.Artist objects matching the specified criteria


class get_artist_list_by_genre(threading.Thread):
    def __init__(self, search_term, order_by, num_results, callback):
        threading.Thread.__init__(self)
        self.search_term = search_term
        self.order_by = order_by
        self.num_results = num_results
        self.callback = callback

    def run(self):
        url = (
            "http://api.jamendo.com/get2/name+id/artist/json/?tag_idstr=%s&order=%s&n=%s"
            % (self.search_term, self.order_by, self.num_results)
        )
        results = get_json(url)
        artists = []
        for result in results:
            item = jamtree.Artist(result['id'], result['name'].strip())
            artists.append(item)
        if artists == []:
            artists = None

        self.callback(artists)


# Gets a list of jamtree.Track objects matching the specified criteria


class get_track_list(threading.Thread):
    def __init__(self, search_term, order_by, num_results, callback):
        threading.Thread.__init__(self)
        self.search_term = search_term
        self.order_by = order_by
        self.num_results = num_results
        self.callback = callback

    def run(self):
        url = (
            "http://api.jamendo.com/get2/id+name+stream+album_id+album_name/"
            "track/json/?searchquery=%s&order=%s&n=%s&streamencoding=ogg2"
            % (self.search_term, self.order_by, self.num_results)
        )
        # print('get_track_list: %s' % url)
        tracks = get_json(url)
        track_list = []
        for track in tracks:
            item = jamtree.Track(track['id'], track['name'].strip(), track['stream'])
            item.album_name = track['album_name']
            track_list.append(item)
        if track_list == []:
            track_list = None

        self.callback(track_list)


# Gets a list of jamtree.Album objects for the specified jamtree.Artist
class get_albums(threading.Thread):
    def __init__(self, artist, callback, add_to_playlist=False):
        threading.Thread.__init__(self)
        self._artist = artist
        self._callback = callback
        self._add_to_playlist = add_to_playlist

    def run(self):
        url = (
            "http://api.jamendo.com/get2/id+name/album/json/?artist_id=%s"
            % self._artist.id
        )
        # print('get_albums: %s' % url)
        albumresults = get_json(url)
        for albumresult in albumresults:
            item = jamtree.Album(albumresult['id'], albumresult['name'].strip())
            self._artist.add_album(item)

        self._callback(self._artist, self._add_to_playlist)


# Gets a list of jamtree.Track objects for the specified jamtree.Album
class get_tracks(threading.Thread):
    def __init__(self, album, callback, add_to_playlist=False):
        threading.Thread.__init__(self)
        self._album = album
        self._callback = callback
        self._add_to_playlist = add_to_playlist

    def run(self):
        url = (
            "http://api.jamendo.com/get2/id+name+stream/track/json/?album_id=%s&streamencoding=ogg2"
            % self._album.id
        )
        # print('get_tracks: %s' % url)
        tracks = get_json(url)
        for track in tracks:
            item = jamtree.Track(track['id'], track['name'].strip(), track['stream'])
            self._album.add_track(item)
        self._callback(self._album, self._add_to_playlist)


# Gets the URL for an album image based on a track id


def get_album_image_url_from_track(track_id):
    url = (
        "http://api.jamendo.com/get2/album_image/track/json/?id=%s&album_imagesize=400"
        % track_id
    )
    imageurl = get_json(url)
    return "".join(imageurl)



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


from xlgui.preferences import widgets
from xl.nls import gettext as _
from xl import xdg

name = _('Lyrics Viewer')
ui = xdg.get_data_path('ui', 'preferences', 'lyrics.ui')


DEFAULT_FONT = None


class LyricsFontPreference(widgets.FontButtonPreference):
    default = None
    name = 'plugin/lyricsviewer/lyrics_font'

    def __init__(self, preferences, widget):
        self.default = DEFAULT_FONT
        widgets.FontButtonPreference.__init__(self, preferences, widget)


class LyricsFontResetButtonPreference(widgets.FontResetButtonPreference):
    default = None
    name = 'plugin/lyricsviewer/reset_button'
    condition_preference_name = 'plugin/lyricsviewer/lyrics_font'

    def __init__(self, preferences, widget):
        self.default = DEFAULT_FONT
        widgets.FontResetButtonPreference.__init__(self, preferences, widget)



# Copyright (C) 2012 Dustin Spicuzza
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import Gdk
from gi.repository import Gtk

import os

from xl import event, providers, player, settings, xdg

from xl.nls import gettext as _

from xlgui import main
from xlgui.widgets import menu, playback

from . import previewprefs

import logging

logger = logging.getLogger(__name__)


class SecondaryOutputPlugin:
    '''Implements logic for plugin'''

    __play_image = Gtk.Image.new_from_icon_name(
        'media-playback-start', Gtk.IconSize.BUTTON
    )
    __pause_image = Gtk.Image.new_from_icon_name(
        'media-playback-pause', Gtk.IconSize.BUTTON
    )

    def get_preferences_pane(self):
        return previewprefs

    def enable(self, exaile):
        self.exaile = exaile

    def on_gui_loaded(self):

        self.hooked = False
        self.resuming = False

        #
        # Initialize the player objects needed
        #

        self.player = player.player.ExailePlayer(
            'preview_device', disable_autoswitch=True
        )
        self.queue = player.queue.PlayQueue(
            self.player,
            location=os.path.join(xdg.get_data_dir(), 'preview_device_queue.state'),
            name='Preview Device Queue',
        )

        #
        # Initialize the GUI stuff
        #

        self._init_gui()

        # preserve state
        if settings.get_option('plugin/previewdevice/shown', True):
            self._init_gui_hooks()

    def disable(self, exaile):
        logger.debug('Disabling Preview Device')
        event.log_event('preview_device_disabling', self, None)
        self._destroy_gui_hooks()
        self._destroy_gui()
        self.player.destroy()

        self.player = None
        self.queue = None

        logger.debug('Preview Device Disabled')

    def _init_gui(self):
        self.pane = Gtk.Paned()

        # stolen from main
        self.info_area = main.MainWindowTrackInfoPane(self.player)
        self.info_area.set_auto_update(True)
        self.info_area.set_border_width(3)
        self.info_area.hide()
        self.info_area.set_no_show_all(True)

        volume_control = playback.VolumeControl(self.player)
        self.info_area.get_action_area().pack_end(volume_control, False, False, 0)

        self.playpause_button = Gtk.Button()
        self.playpause_button.set_relief(Gtk.ReliefStyle.NONE)
        self._on_playback_end(None, None, None)
        self.playpause_button.connect(
            'button-press-event', self._on_playpause_button_clicked
        )

        self.progress_bar = playback.SeekProgressBar(self.player, use_markers=False)
        self.progress_bar.set_valign(Gtk.Align.CENTER)

        play_toolbar = Gtk.Box()
        play_toolbar.pack_start(self.playpause_button, False, False, 0)
        play_toolbar.pack_start(self.progress_bar, True, True, 0)
        play_toolbar.child_set_property(self.progress_bar, 'padding', 3)

        # stick our player controls into this box
        self.pane1_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        self.pane2_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.pane2_box.pack_start(self.info_area, False, False, 0)
        self.pane2_box.pack_start(play_toolbar, False, False, 0)

        self.pane.pack1(self.pane1_box, resize=True, shrink=True)
        self.pane.pack2(self.pane2_box, resize=True, shrink=True)

        # setup menus

        # add menu item to 'view' to display our playlist
        self.menu = menu.check_menu_item(
            'preview_player',
            '',
            _('Preview Player'),
            lambda *e: self.hooked,
            self._on_view,
        )

        providers.register('menubar-view-menu', self.menu)

        self.preview_menuitem = menu.simple_menu_item(
            '_preview',
            ['enqueue'],
            _('Preview'),
            callback=self._on_preview,
            condition_fn=lambda n, p, c: not c['selection-empty'],
        )

        # TODO: Setup on other context menus
        self.preview_provides = ['track-panel-menu', 'playlist-context-menu']

        for provide in self.preview_provides:
            providers.register(provide, self.preview_menuitem)

        self._on_option_set('gui_option_set', settings, 'gui/show_info_area')
        self._on_option_set('gui_option_set', settings, 'gui/show_info_area_covers')
        event.add_ui_callback(self._on_option_set, 'option_set')

    def _destroy_gui(self):
        event.remove_callback(self._on_option_set, 'option_set')

        for provide in self.preview_provides:
            providers.unregister(provide, self.preview_menuitem)
        providers.unregister('menubar-view-menu', self.menu)

        self.info_area.destroy()
        self.playpause_button.destroy()

        self.pane2_box.destroy()
        self.pane1_box.destroy()
        self.pane.destroy()

    def _setup_events(self, setup):
        setup(self._on_playback_resume, 'playback_player_resume', self.player)
        setup(self._on_playback_end, 'playback_player_end', self.player)
        setup(self._on_playback_error, 'playback_error', self.player)
        setup(self._on_playback_start, 'playback_track_start', self.player)
        setup(self._on_toggle_pause, 'playback_toggle_pause', self.player)

    def _init_gui_hooks(self):
        """
        Initializes any hooks into the main Exaile GUI

        Note that this is rather ugly, but currently exaile doesn't really
        have a better way to do this, and there isn't a better place to
        stick our gui objects.
        """

        if self.hooked:
            return

        # the info_area will be where we sit, and the info_area
        # will be duplicated for two sides

        # need to move the play_toolbar, and duplicate it
        # also once for each side

        info_area = main.mainwindow().info_area
        play_toolbar = main.mainwindow().builder.get_object('play_toolbar')

        parent = play_toolbar.get_parent()
        parent.remove(play_toolbar)

        parent = info_area.get_parent()
        parent.remove(info_area)

        parent.pack_start(self.pane, False, False, 0)
        parent.reorder_child(self.pane, 0)

        # stick the main player controls into this box
        self.pane1_box.pack_start(info_area, False, False, 0)
        self.pane1_box.pack_start(play_toolbar, False, False, 0)

        # and do it
        self.pane.show_all()

        # add player events
        self._setup_events(event.add_ui_callback)

        self.hooked = True
        settings.set_option('plugin/previewdevice/shown', True)

        logger.debug("Preview device gui hooked")
        event.log_event('preview_device_enabled', self, None)

    def _destroy_gui_hooks(self):
        """
        Removes any hooks from the main Exaile GUI
        """

        if not self.hooked:
            return

        info_area = main.mainwindow().info_area
        play_toolbar = main.mainwindow().builder.get_object('play_toolbar')

        # detach main GUI elements
        parent = play_toolbar.get_parent()
        parent.remove(play_toolbar)

        parent = info_area.get_parent()
        parent.remove(info_area)

        # detach the element we added to hold them
        parent = self.pane.get_parent()
        parent.remove(self.pane)

        # reattach
        parent.pack_start(info_area, False, False, 0)
        parent.reorder_child(info_area, 0)
        parent.pack_start(play_toolbar, False, False, 0)

        # remove player events
        self._setup_events(event.remove_callback)

        self.hooked = False
        settings.set_option('plugin/previewdevice/shown', False)
        logger.debug('Preview device unhooked')

    #
    # Menu events
    #

    def _on_view(self, menu, name, parent, context):
        if self.hooked:
            self._destroy_gui_hooks()
        else:
            self._init_gui_hooks()

    def _on_preview(self, menu, display_name, playlist_view, context):
        self._init_gui_hooks()
        tracks = context['selected-tracks']
        if len(tracks) > 0:
            self.queue.play(tracks[0])

    #
    # Various player events
    #

    def _on_playpause_button_clicked(self, widget, event):
        """
        Called when the play button is clicked
        """

        if event.button == Gdk.BUTTON_PRIMARY:
            if event.type == Gdk.EventType.BUTTON_PRESS and (
                self.player.is_paused() or self.player.is_playing()
            ):
                self.player.toggle_pause()
            elif event.type == Gdk.EventType._2BUTTON_PRESS:
                self.player.stop()

    def _on_option_set(self, name, object, option):
        """
        Handles changes of settings
        """
        if option == 'gui/show_info_area':
            self.info_area.set_no_show_all(False)
            if settings.get_option(option, True):
                self.info_area.show_all()
            else:
                self.info_area.hide()
            self.info_area.set_no_show_all(True)

        elif option == 'gui/show_info_area_covers':
            cover = self.info_area.cover
            cover.set_no_show_all(False)
            if settings.get_option(option, True):
                cover.show_all()
            else:
                cover.hide()
            cover.set_no_show_all(True)

    def _on_playback_resume(self, type, player, data):
        self.resuming = True

    def _on_playback_start(self, type, player, object):
        """
        Called when playback starts
        Sets the currently playing track visible in the currently selected
        playlist if the user has chosen this setting
        """
        if self.resuming:
            self.resuming = False
            return

        self.playpause_button.set_image(self.__pause_image)
        self.playpause_button.set_tooltip_text(
            _('Pause Playback (double click to stop)')
        )

    def _on_playback_end(self, type, player, object):
        """
        Called when playback ends
        """
        self.playpause_button.set_image(self.__play_image)
        self.playpause_button.set_tooltip_text(_('Start Playback'))

    def _on_playback_error(self, type, player, message):
        """
        Called when there has been a playback error
        """
        main.mainwindow().message.show_error(_('Playback error encountered!'), message)

    def _on_toggle_pause(self, type, player, object):
        """
        Called when the user clicks the play button after playback has
        already begun
        """
        if player.is_paused():
            image = self.__play_image
            tooltip = _('Continue Playback')
        else:
            image = self.__pause_image
            tooltip = _('Pause Playback')

        self.playpause_button.set_image(image)
        self.playpause_button.set_tooltip_text(tooltip)


plugin_class = SecondaryOutputPlugin



from xlgui.preferences import widgets
from xl.nls import gettext as _
import os

name = _('DAAP Client')
basedir = os.path.dirname(os.path.realpath(__file__))
ui = os.path.join(basedir, 'daapclient_prefs.ui')


class Ipv6Preference(widgets.CheckPreference):
    default = False
    name = 'plugin/daapclient/ipv6'


class HistoryPreference(widgets.CheckPreference):
    default = True
    name = 'plugin/daapclient/history'



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


import time


def timeit(n, func, *args):
    runs = []
    for n in range(0, n):
        start = time.clock()
        func.__call__(*args)
        runs.append(time.clock() - start)
    avg = sum(runs) / len(runs)
    print("%1.8f seconds" % avg)



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import hashlib
import logging
import os
from typing import Any, Callable, Optional

from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Pango

from xl.nls import gettext as _
from xl import event, main, settings
from xlgui import guiutil
from xlgui.widgets import dialogs
from xlgui.guiutil import GtkTemplate

logger = logging.getLogger(__name__)


class Preference:
    """
    Representing a Gtk.Entry preferences item
    """

    default: Any = ''
    done: Callable[[], bool]
    label_widget: Optional[Gtk.Widget]
    name: str
    restart_required = False

    def __init__(self, preferences, widget):
        """
        Initializes the preferences item
        expects the name of the widget in the designer file, the default for
        this setting, an optional function to be called when the value is
        changed, and an optional function to be called when this setting
        is applied
        """

        self.widget = widget
        self.preferences = preferences

        if self.restart_required:
            self.message = dialogs.MessageBar(
                parent=preferences.builder.get_object('preferences_box'),
                type=Gtk.MessageType.QUESTION,
                buttons=Gtk.ButtonsType.CLOSE,
                text=_('Restart Exaile?'),
            )
            self.message.set_secondary_text(
                _('A restart is required for this change to take effect.')
            )

            button = self.message.add_button(_('Restart'), Gtk.ResponseType.ACCEPT)
            button.set_image(
                Gtk.Image.new_from_icon_name('view-refresh', Gtk.IconSize.BUTTON)
            )

            self.message.connect('response', self.on_message_response)

        self._set_value()
        self._setup_change()

    def change(self, *args):
        self.apply()

    def _setup_change(self):
        """
        Sets up the function to be called when this preference is changed
        """
        self.widget.connect(
            'focus-out-event', self.change, self.name, self._get_value()
        )

        try:
            self.widget.connect(
                'activate',
                lambda *e: self.change(self.widget, None, self.name, self._get_value()),
            )
        except TypeError:
            pass

    def _get_value(self):
        """
        Value to be stored into the settings file
        """
        return self.widget.get_text()

    def _set_value(self):
        """
        Sets the GUI widget up for this preference
        """
        if not self.widget:
            logger.error("Widget not found: %s", self.name)
            return
        self.widget.set_text(str(settings.get_option(self.name, self.default)))

    def apply(self, value=None):
        """
        Applies this setting
        """
        if hasattr(self, 'done') and not self.done():
            return False

        oldvalue = settings.get_option(self.name, self.default)

        if value is None:
            value = self._get_value()

        if value != oldvalue:
            settings.set_option(self.name, value)

            if self.restart_required:
                self.message.show()

        return True

    def on_message_response(self, widget, response):
        """
        Restarts Exaile if requested
        """
        widget.hide()

        if response == Gtk.ResponseType.ACCEPT:
            GLib.idle_add(main.exaile().quit, True)

    def hide_widget(self):
        '''Hides the widget and optionally its associated label'''
        self.widget.hide()
        if hasattr(self, 'label_widget'):
            self.label_widget.hide()

    def show_widget(self):
        '''Shows the widget and optionally its associated label'''
        self.widget.show_all()
        if hasattr(self, 'label_widget'):
            self.label_widget.show_all()

    def set_widget_sensitive(self, value):
        '''Sets sensitivity of widget and optionally its associated label'''
        self.widget.set_sensitive(value)
        if hasattr(self, 'label_widget'):
            self.label_widget.set_sensitive(value)


class Conditional:
    """
    Allows for reactions on changes
    of other preference items
    """

    condition_preference_name = ''
    condition_widget = None

    def __init__(self):
        event.add_ui_callback(self.on_option_set, 'option_set')
        GLib.idle_add(
            self.on_option_set, 'option_set', settings, self.condition_preference_name
        )

    def get_condition_value(self):
        """
        :returns: The currently selected value in the condition widget,
                  presumes it is a combo box
        """
        i = self.condition_widget.get_active_iter()
        return self.condition_widget.get_model().get_value(i, 0)

    def on_check_condition(self):
        """
        Specifies the condition to meet

        :returns: Whether the condition is met or not
        :rtype: bool
        """
        pass

    def on_condition_met(self):
        """
        Called as soon as the
        specified condition is met
        """
        self.widget.set_sensitive(True)

    def on_condition_failed(self):
        """
        Called as soon as the specified
        condition is not met anymore
        """
        self.widget.set_sensitive(False)

    def on_option_set(self, event, settings, option):
        """
        Called as soon as options change
        """
        if option == self.condition_preference_name:
            if self.on_check_condition():
                self.on_condition_met()
            else:
                self.on_condition_failed()


class CheckConditional(Conditional):
    """
    True if the conditional widget is active
    """

    def get_condition_value(self):
        return self.condition_widget.get_active()

    def on_check_condition(self):
        """
        Specifies the condition to meet

        :returns: Whether the condition is met or not
        :rtype: bool
        """
        return self.get_condition_value()


class MultiConditional:
    """
    Allows for reactions on changes of multiple preference items
    """

    condition_preference_names = []
    condition_widgets = {}

    def __init__(self):
        event.add_ui_callback(self.on_option_set, 'option_set')
        GLib.idle_add(
            self.on_option_set,
            'option_set',
            settings,
            self.condition_preference_names[0],
        )

    def get_condition_value(self, name):
        """
        :returns: The currently selected value in the condition widget,
                  presumes it is a combo box
        """
        widget = self.condition_widgets[name]
        return widget.get_model().get_value(widget.get_active_iter(), 0)

    def on_check_condition(self):
        """
        Specifies the condition to meet

        :returns: Whether the condition is met or not
        :rtype: bool
        """
        pass

    def on_condition_met(self):
        """
        Called as soon as the
        specified condition is met
        """
        self.widget.set_sensitive(True)

    def on_condition_failed(self):
        """
        Called as soon as the specified
        condition is not met anymore
        """
        self.widget.set_sensitive(False)

    def on_option_set(self, event, settings, option):
        """
        Called as soon as options change
        """
        if option in self.condition_preference_names:
            if self.on_check_condition():
                self.on_condition_met()
            else:
                self.on_condition_failed()


class Button(Preference):
    """
    Represents a button for custom usage
    """

    def __init__(self, preferences, widget):
        Preference.__init__(self, preferences, widget)

        widget.connect('clicked', self.on_clicked)

    def _setup_change(self, *e):
        pass

    def _get_value(self):
        return None

    def _set_value(self):
        pass

    def apply(self, *e):
        return False

    def on_clicked(self, button):
        """ Override """
        pass


class HashedPreference(Preference):
    """
    Represents a text entry with automated hashing

    Options:
    * type (Which hashfunction to use, default: md5)
    """

    type = 'md5'

    def __init__(self, preferences, widget):
        Preference.__init__(self, preferences, widget)

        self.widget.set_visibility(True)
        self._delete_text_id = self.widget.connect('delete-text', self.on_delete_text)
        self._insert_text_id = self.widget.connect('insert-text', self.on_insert_text)

    def _setup_change(self):
        """
        Sets up the function to be called when this preference is changed
        """
        self.widget.connect('focus-out-event', lambda *e: self.apply())

    def done(self):
        """
        Determines if changes are to be expected
        """
        if self._delete_text_id is None and self._insert_text_id is None:
            return True

        return False

    def apply(self, value=None):
        """
        Applies this setting
        """
        if not self.done():
            return False

        if value is None:
            value = self._get_value()
        if value is None:
            return True

        if value != '':
            hashfunc = hashlib.new(self.type)
            hashfunc.update(value.encode('utf-8'))
            value = hashfunc.hexdigest()

        oldvalue = settings.get_option(self.name, self.default)

        if value != oldvalue:
            settings.set_option(self.name, value)

        self.widget.set_text(value)
        self.widget.set_visibility(True)
        self._delete_text_id = self.widget.connect('delete-text', self.on_delete_text)
        self._insert_text_id = self.widget.connect('insert-text', self.on_insert_text)

        return True

    def on_delete_text(self, widget, start, end):
        """
        Clears the text entry and makes following input invisible
        """
        self.widget.disconnect(self._delete_text_id)
        self.widget.disconnect(self._insert_text_id)
        self._delete_text_id = self._insert_text_id = None

        self.widget.set_visibility(False)
        self.widget.set_text('')

    def on_insert_text(self, widget, text, length, position):
        """
        Clears the text entry and makes following input invisible
        """
        self.widget.disconnect(self._delete_text_id)
        self.widget.disconnect(self._insert_text_id)
        self._delete_text_id = self._insert_text_id = None

        self.widget.set_visibility(False)
        # Defer to after returning from this method
        GLib.idle_add(self.widget.set_text, text)
        GLib.idle_add(self.widget.set_position, length)


class CheckPreference(Preference):
    """
    A class to represent check boxes in the preferences window
    """

    def __init__(self, preferences, widget):
        Preference.__init__(self, preferences, widget)

    def _setup_change(self):
        self.widget.connect('toggled', self.change)

    def _set_value(self):
        self.widget.set_active(settings.get_option(self.name, self.default))

    def _get_value(self):
        return self.widget.get_active()


class DirPreference(Preference):
    """
    Directory chooser button
    """

    def __init__(self, preferences, widget):
        Preference.__init__(self, preferences, widget)

    def _setup_change(self):
        self.widget.connect('current-folder-changed', self.change)

    def _set_value(self):
        """
        Sets the current directory
        """
        directory = os.path.expanduser(settings.get_option(self.name, self.default))
        if not os.path.exists(directory):
            os.makedirs(directory)
        self.widget.set_current_folder(directory)

    def _get_value(self):
        return self.widget.get_filename()


class OrderListPreference(Preference):
    """
    A list box with reorderable items
    """

    def __init__(self, preferences, widget):
        self.model = Gtk.ListStore(str)
        Preference.__init__(self, preferences, widget)
        widget.set_headers_visible(False)
        widget.set_reorderable(True)

        text = Gtk.CellRendererText()
        col = Gtk.TreeViewColumn("Item", text, text=0)
        self.widget.append_column(col)
        self.widget.set_model(self.model)

    def _setup_change(self):
        self.widget.connect('drag-end', self.change)

    def _set_value(self):
        """
        Sets the preferences for this widget
        """
        items = settings.get_option(self.name, self.default)

        self.model.clear()
        for item in items:
            self.model.append([item])

    def _get_value(self):
        """
        Value to be stored into the settings file
        """
        items = []

        for row in self.model:
            items += [row[0]]

        return items


class SelectionListPreference(Preference):
    """
    A list allowing for enabling/disabling
    as well as reordering of items

    Options:
    * items: list of :class:`SelectionListPreference.Item` objects
    * default: list of item ids
    """

    class Item:
        """
        Convenience class for preference item description
        """

        def __init__(self, id, title, description=None, fixed=False):
            """
            :param id: the unique identifier
            :type id: string
            :param title: the readable title
            :type title: string
            :param description: optional description of the item
            :type description: string
            :param fixed: whether the item should be removable
            :type fixed: bool
            """
            self.__id = id
            self.__title = title
            self.__description = description
            self.__fixed = fixed

        id = property(lambda self: self.__id)
        title = property(lambda self: self.__title)
        description = property(lambda self: self.__description)
        fixed = property(lambda self: self.__fixed)

    @GtkTemplate('ui', 'preferences', 'widgets', 'selection_list_preference.ui')
    class InternalWidget(Gtk.ScrolledWindow):
        """
        Internal class for making GtkTemplate work with subclassing
        """

        __gtype_name__ = 'InternalWidget'

        (
            model,
            tree,
            toggle_renderer,
            text_renderer,
            enabled_column,
            title_column,
        ) = GtkTemplate.Child.widgets(6)
        selectionlp = None

        def __init__(self, preference):
            Gtk.ScrolledWindow.__init__(self)
            self.init_template()
            self.selectionlp = preference

            self.tree.enable_model_drag_source(
                Gdk.ModifierType.BUTTON1_MASK,
                [('GTK_TREE_MODEL_ROW', Gtk.TargetFlags.SAME_WIDGET, 0)],
                Gdk.DragAction.MOVE,
            )
            self.tree.enable_model_drag_dest(
                [('GTK_TREE_MODEL_ROW', Gtk.TargetFlags.SAME_WIDGET, 0)],
                Gdk.DragAction.MOVE,
            )
            self.tree.connect('drag-end', self.selectionlp.change)

            self.enabled_column.set_cell_data_func(
                self.toggle_renderer, self.enabled_data_function
            )

            self.title_column.set_cell_data_func(
                self.text_renderer, self.title_data_function
            )

        @GtkTemplate.Callback
        def on_row_activated(self, tree, path, column):
            """
            Updates the enabled column
            """
            if self.model[path][4]:
                return

            enabled = not self.model[path][3]
            self.model[path][3] = enabled

        def enabled_data_function(self, column, cell, model, iter, user_data):
            """
            Prepares sensitivity
            of the enabled column
            """
            path = model.get_path(iter)
            fixed = model[path][4]
            cell.props.sensitive = not fixed

        def title_data_function(self, column, cell, model, iter, user_data):
            """
            Prepares the markup to be
            used for the title column
            """
            path = model.get_path(iter)
            title, description = model[path][1], model[path][2]

            markup = '<b>%s</b>' % title

            if description is not None:
                markup += '\n<span size="small">%s</span>' % description

            cell.props.markup = markup

        def iter_prev(self, iter, model):
            """
            Returns the previous iter
            Taken from PyGtk FAQ 13.51
            """
            path = model.get_path(iter)
            position = path[-1]

            if position == 0:
                return None

            prev_path = list(path)[:-1]
            prev_path.append(position - 1)
            prev = model.get_iter(tuple(prev_path))

            return prev

        @GtkTemplate.Callback
        def on_key_press_event(self, tree, event):
            """
            Allows for reordering via keyboard (Alt+<direction>)
            """
            if not event.get_state() & Gdk.ModifierType.MOD1_MASK:
                return

            if event.keyval not in (Gdk.KEY_Up, Gdk.KEY_Down):
                return

            model, selected_iter = tree.get_selection().get_selected()

            if event.keyval == Gdk.KEY_Up:
                previous_iter = self.iter_prev(selected_iter, model)
                model.move_before(selected_iter, previous_iter)
            elif event.keyval == Gdk.KEY_Down:
                next_iter = model.iter_next(selected_iter)
                model.move_after(selected_iter, next_iter)

            tree.scroll_to_cell(model.get_path(selected_iter))

            self.selectionlp.apply()

        @GtkTemplate.Callback
        def on_toggled(self, cell, path):
            """
            Updates the enabled column
            """
            if self.model[path][4]:
                return

            active = not cell.get_active()
            cell.set_active(active)
            self.model[path][3] = active

            self.selectionlp.apply()

    def __init__(self, preferences, widget):
        internal_widget = self.InternalWidget(self)
        self.model = internal_widget.model

        for item in self.items:
            row = [item.id, item.title, item.description, True, item.fixed]
            self.model.append(row)

        guiutil.gtk_widget_replace(widget, internal_widget)
        Preference.__init__(self, preferences, internal_widget)

    def _get_value(self):
        """
        Value to be stored in the settings
        """
        return [row[0] for row in self.model if row[3]]

    def _set_value(self):
        """
        Updates the internal representation
        """
        selected_items = settings.get_option(self.name, self.default)
        # Get list of available items
        available_items = [row[0] for row in self.model]

        if not available_items:
            return

        # Filter out invalid items
        selected_items = [item for item in selected_items if item in available_items]
        # Cut out unselected items
        unselected_items = [
            item for item in available_items if item not in selected_items
        ]
        # Move unselected items to the end
        items = selected_items + unselected_items
        new_order = [available_items.index(item) for item in items]
        self.model.reorder(new_order)

        # Disable unselected items
        for row in self.model:
            if row[0] in unselected_items and not row[4]:
                row[3] = False
            else:
                row[3] = True


class ShortcutListPreference(Preference):
    """
    A list showing available items and allowing
    to assign/edit/remove key accelerators
    """

    def __init__(self, preferences, widget):
        self.list = Gtk.ListStore(str, str)

        Preference.__init__(self, preferences, widget)

        self.widget.set_model(self.list)

        title_renderer = Gtk.CellRendererText()
        title_column = Gtk.TreeViewColumn(_('Action'), title_renderer, text=0)
        title_column.set_expand(True)
        title_column.set_cell_data_func(title_renderer, self.title_data_func)
        accel_renderer = Gtk.CellRendererAccel()
        accel_renderer.set_property('editable', True)
        accel_renderer.set_property('style', Pango.Style.OBLIQUE)
        accel_renderer.connect('accel-cleared', self.on_accel_cleared)
        accel_renderer.connect('accel-edited', self.on_accel_edited)
        accel_column = Gtk.TreeViewColumn(_('Shortcut'), accel_renderer, text=1)
        accel_column.set_expand(True)

        self.widget.append_column(title_column)
        self.widget.append_column(accel_column)

    def title_data_func(self, celllayout, cell, model, iter, user_data):
        """
        Renders human readable titles instead of the actual keys
        """
        key = model.get_value(iter, 0)

        try:
            cell.set_property('text', self.available_items[key])
        except KeyError:
            pass

    def on_accel_cleared(self, cellrenderer, path):
        """
        Clears accelerators in the list
        """
        iter = self.list.get_iter(path)
        self.list.set_value(iter, 1, '')

    def on_accel_edited(self, cellrenderer, path, accel_key, accel_mods, keycode):
        """
        Updates accelerators display in the list
        """
        accel = Gtk.accelerator_name(accel_key, accel_mods)
        iter = self.list.get_iter(path)
        self.list.set_value(iter, 1, accel)

    def _set_value(self):
        """
        Sets the preferences for this widget
        """
        items = settings.get_option(self.name, self.default)
        self.update_list(items)

    def _get_value(self):
        """
        Value to be stored into the settings file
        """
        option = {}

        iter = self.list.get_iter_first()
        while iter:
            action = self.list.get_value(iter, 0)
            accel = self.list.get_value(iter, 1)
            if accel:
                option[action] = accel
            iter = self.list.iter_next(iter)

        return option

    def update_list(self, items):
        """
        Updates the displayed items
        """
        self.list.clear()
        for action in self.available_items.keys():
            try:
                accel = items[action]
            except KeyError:
                accel = ''
            self.list.append([action, accel])


class TextViewPreference(Preference):
    """
    Represents a Gtk.TextView
    """

    def __init__(self, preferences, widget):
        """
        Initializes the object
        """
        Preference.__init__(self, preferences, widget)

    def _setup_change(self):
        self.widget.connect('focus-out-event', self.change)

    def get_all_text(self):
        """
        Returns the value of the text buffer
        """
        buf = self.widget.get_buffer()
        start = buf.get_start_iter()
        end = buf.get_end_iter()
        return buf.get_text(start, end, True)

    def _set_value(self):
        """
        Sets the value of this widget
        """
        self.widget.get_buffer().set_text(
            str(settings.get_option(self.name, default=self.default))
        )

    def _get_value(self):
        """
        Applies the setting
        """
        return self.get_all_text()


class ListPreference(Preference):
    """
    A class to represent a space separated list in the preferences window
    """

    def __init__(self, preferences, widget):
        Preference.__init__(self, preferences, widget)

    def _set_value(self):
        items = settings.get_option(self.name, default=self.default)
        try:
            items = " ".join(items)
        except TypeError:
            items = ""
        self.widget.set_text(items)

    def _get_value(self):
        import shlex

        return shlex.split(self.widget.get_text())


class SpinPreference(Preference):
    """
    A class to represent a numeric entry box with stepping buttons
    """

    def __init__(self, preferences, widget):
        Preference.__init__(self, preferences, widget)

    def _set_value(self):
        value = settings.get_option(self.name, default=self.default)
        self.widget.set_value(value)

    def _setup_change(self):
        self.widget.connect('value-changed', self.change)

    def _get_value(self):
        return self.widget.get_value()


class ScalePreference(SpinPreference):
    """
    Representation of Gtk.Scale widgets
    """

    def __init__(self, preferences, widget):
        SpinPreference.__init__(self, preferences, widget)


class FloatPreference(Preference):
    """
    A class to represent a floating point number in the preferences window
    """

    def __init__(self, preferences, widget):
        Preference.__init__(self, preferences, widget)

    def _set_value(self):
        self.widget.set_text(str(settings.get_option(self.name, default=self.default)))

    def _get_value(self):
        return float(self.widget.get_text())


class IntPreference(FloatPreference):
    def _get_value(self):
        return int(self.widget.get_text())


class RGBAButtonPreference(Preference):
    """
    A class to represent the color button
    """

    def __init__(self, preferences, widget):
        Preference.__init__(self, preferences, widget)

    def _setup_change(self):
        self.widget.connect('color-set', self.change)

    def _set_value(self):
        value = settings.get_option(self.name, self.default)
        rgba = Gdk.RGBA()
        rgba.parse(value)
        self.widget.set_rgba(rgba)

    def _get_value(self):
        return self.widget.get_rgba().to_string()


class FontButtonPreference(Preference):
    """
    Font button
    """

    def __init__(self, preferences, widget):
        Preference.__init__(self, preferences, widget)

    def _setup_change(self):
        self.widget.connect('font-set', self.change)

    def _set_value(self):
        font = settings.get_option(self.name, self.default)
        self.widget.set_font_name(font)

    def _get_value(self):
        font = self.widget.get_font_name()
        return font


class FontResetButtonPreference(Button, Conditional):
    """
    A button to reset a font button to a default font
    """

    def __init__(self, preferences, widget):
        Button.__init__(self, preferences, widget)
        Conditional.__init__(self)

    def on_check_condition(self):
        if self.condition_widget.get_font_name() == self.default:
            return False
        return True

    def on_clicked(self, button):
        self.condition_widget.set_font_name(self.default)
        self.condition_widget.emit('font-set')


class ComboPreference(Preference):
    """
    A combo box. The value stored in the settings must be the
    first column of the combo box model.
    """

    def __init__(self, preferences, widget):
        Preference.__init__(self, preferences, widget)

    def _setup_change(self):
        self.widget.connect('changed', self.change)

    def _set_value(self):
        """
        Sets the preferences for this widget
        """
        item = settings.get_option(self.name, self.default)

        model = self.widget.get_model()

        for row in model:
            if item == row[0]:
                self.widget.set_active_iter(row.iter)

    def _get_value(self):
        """
        Value to be stored into the settings file
        """
        model = self.widget.get_model()
        iter = self.widget.get_active_iter()

        return model.get_value(iter, 0)


class ComboEntryPreference(Preference):
    """
    A combo box allowing for user defined
    values, presets and auto completion

    Options:
    * completion_items (List of completion items or
      dictionary of items and their titles)
    * preset_items (List of preset items or
      dictionary of items and their titles)
    """

    def __init__(self, preferences, widget):
        Preference.__init__(self, preferences, widget)

        self.list = Gtk.ListStore(str)

        try:
            try:
                preset_items = list(self.preset_items.items())
                self.list = Gtk.ListStore(str, str)
                text_renderer = self.widget.get_cells()[0]
                text_renderer.set_property('weight', Pango.Weight.BOLD)

                title_renderer = Gtk.CellRendererText()
                self.widget.pack_start(title_renderer, False)
                self.widget.add_attribute(title_renderer, 'text', 1)
            except AttributeError:
                preset_items = [[item] for item in self.preset_items]

            for preset in preset_items:
                self.list.append(preset)
        except AttributeError:
            pass

        self.widget.set_model(self.list)
        self.widget.set_entry_text_column(0)

        try:
            completion = Gtk.EntryCompletion()

            try:
                completion_items = list(self.completion_items.items())
                self.completion_list = Gtk.ListStore(str, str)

                title_renderer = Gtk.CellRendererText()
                completion.pack_end(title_renderer, True)
                completion.add_attribute(title_renderer, 'text', 1)
            except AttributeError:
                completion_items = [[item] for item in self.completion_items]
                self.completion_list = Gtk.ListStore(str)

            keyword_renderer = Gtk.CellRendererText()
            keyword_renderer.set_property('weight', Pango.Weight.BOLD)
            completion.pack_end(keyword_renderer, True)
            completion.add_attribute(keyword_renderer, 'text', 0)
            completion.set_match_func(self.on_matching)
            completion.connect('match-selected', self.on_match_selected)

            completion.set_popup_single_match(True)
            completion.set_model(self.completion_list)
            self.widget.get_child().set_completion(completion)

            for item in completion_items:
                self.completion_list.append(item)
        except AttributeError:
            pass

    def _setup_change(self):
        """
        Sets up the function to be called
        when this preference is changed
        """
        self.widget.connect('changed', self.change, self.name, self._get_value())

    def _set_value(self):
        """
        Sets the preferences for this widget
        """
        value = settings.get_option(self.name, self.default)
        self.widget.get_child().set_text(str(value))

    def _get_value(self):
        """
        Value to be stored into the settings file
        """
        return self.widget.get_child().get_text()

    def on_matching(self, completion, text, iter):
        """
        Matches the content of this box to
        the list of available completions
        """
        cursor_pos = self.widget.get_child().get_position()
        # Ignore the rest, allows for completions in the middle
        text = text[:cursor_pos]
        match = self.completion_list.get_value(iter, 0)

        # Try to find match, from largest to smallest
        for i in range(len(match), -1, -1):
            # Find from the rear
            match_pos = text.rfind(match[:i])
            # Matched if the match is not empty
            # and equal to the text from the matched position to the end
            # and not equal to the match itself
            # (the latter hides the match if it was already fully typed)
            if match[:i] and match[:i] == text[match_pos:] and match[:i] != match:
                return True
        return False

    def on_match_selected(self, completion, list, iter):
        """
        Inserts the selected completion
        """
        cursor_pos = self.widget.get_child().get_position()
        text = self.widget.get_child().get_text()[:cursor_pos]
        match = list.get_value(iter, 0)

        for i in range(len(match), -1, -1):
            match_pos = text.rfind(match[:i])
            if match[:i] and match[:i] == text[match_pos:]:
                # Delete halfway typed text
                self.widget.get_child().delete_text(
                    match_pos, match_pos + len(match[:i])
                )
                # Insert match at matched position
                self.widget.get_child().insert_text(match, match_pos)
                # Update cursor position
                self.widget.get_child().set_position(match_pos + len(match))

        return True


# vim: et sts=4 sw=4



# Copyright (C) 2014 Dustin Spicuzza
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import logging

from xl.nls import gettext as _
from xl import providers, settings

from xlgui.widgets import menu, notebook
from xlgui.panel import lyrics


logger = logging.getLogger(__name__)


class PanelData:

    __slots__ = ['tab', 'menuitem', 'panel', 'position', 'shown']

    def __init__(self, tab, panel, position, menuitem):
        self.tab = tab  # Notebook tab
        self.menuitem = menuitem  # Menuitem
        self.panel = panel  # Panel provider
        self.position = position  # Position in notebook
        self.shown = True  # Whether the panel is shown

    @property
    def opts(self):
        return (self.shown, self.position)

    def __repr__(self):
        return '<PanelData tab: %s, panel: %s, position: %s, shown: %s>' % (
            self.tab,
            self.panel,
            self.position,
            self.shown,
        )


class PanelNotebook(notebook.SmartNotebook, providers.ProviderHandler):
    """
    This notebook holds the panels shown on the left side of the main
    UI. Do not directly add things to this panel, but instead register
    a provider.

    The provider object must have the following attributes:

        name: the name of the provider

    The provider object must have the following methods:

        get_panel: This must return a widget that is derived from
                   xlgui.widgets.notebook.NotebookPage
    """

    def __init__(self, exaile, gui):
        notebook.SmartNotebook.__init__(self, vertical=True)

        self.exaile = exaile
        self.panels = {}  # key: name, value: PanelData object

        self.set_add_tab_on_empty(False)

        self.loading_panels = True

        self.connect('page-removed', self.on_panel_removed)
        self.connect('page-reordered', self.on_panel_reordered)
        self.connect('switch-page', self.on_panel_switch)

        _register_builtin_panels(exaile, gui.main.window)

        self.view_menu = menu.ProviderMenu('panel-tab-context', None)

        # setup/register the view menu
        menu.simple_menu_item(
            'panel-menu', ['show-playing-track'], _('P_anels'), submenu=self.view_menu
        ).register('menubar-view-menu')

        providers.ProviderHandler.__init__(self, 'main-panel', simple_init=True)

        # Provide interface for adding buttons to the notebook
        self.actions = notebook.NotebookActionService(self, 'main-panel-actions')

        if not self.exaile.loading:
            self.on_gui_loaded()

    def focus_panel(self, tab_name):

        data = self.panels[tab_name]
        if data.shown:
            panel_nr = self.page_num(data.tab.page)
            self.set_current_page(panel_nr)
            data.tab.grab_focus()

    def get_active_panel(self):
        self.get_current_page()
        return None

    def toggle_panel(self, tab_name):

        data = self.panels[tab_name]

        if data.shown:
            self.remove_tab(data.tab)
        else:
            self.add_tab(data.tab, data.tab.page, data.position)
            data.shown = True

            self.save_panel_settings()

    def on_provider_added(self, provider):

        if provider.name is None:
            logger.warning(
                "Ignoring improperly initialized panel provider: %s", provider
            )
            return

        panel = provider.get_panel()
        panel.show()

        tab = notebook.NotebookTab(self, panel, vertical=True)
        tab.provider = provider

        item = menu.check_menu_item(
            provider.name,
            [],
            panel.get_page_name(),
            lambda *a: self.panels[provider.name].shown,
            lambda *a: self.toggle_panel(provider.name),
        )

        providers.register('panel-tab-context', item)

        self.add_tab(tab, panel)
        self.panels[provider.name] = PanelData(
            tab, provider, self.get_n_pages() - 1, item
        )

        self.save_panel_settings()

    def on_provider_removed(self, provider):

        data = self.panels[provider.name]

        for n in range(self.get_n_pages()):
            if data.tab.page == self.get_nth_page(n):
                self.remove_page(n)
                break

        providers.unregister('panel-tab-context', data.menuitem)
        del self.panels[provider.name]

    def on_panel_removed(self, notebook, page, pagenum):

        if self.loading_panels:
            return

        for name, data in self.panels.items():
            if data.tab.page == page:
                data.shown = False
                break

        self.save_panel_settings()

    def on_panel_reordered(self, notebook, page, pagenum):

        if self.loading_panels:
            return

        for name, data in self.panels.items():
            if data.shown:
                data.position = self.page_num(data.tab.page)

        self.save_panel_settings()

    def on_panel_switch(self, notebook, page, pagenum):
        """
        Saves the currently selected panel
        """
        if self.exaile.loading:
            return

        page = notebook.get_nth_page(pagenum)
        for name, data in self.panels.items():
            if data.tab.page == page:
                settings.set_option('gui/last_selected_panel', name)
                return

    def save_panel_settings(self):

        if self.loading_panels:
            return

        param = dict([(k, v.opts) for k, v in self.panels.items()])
        settings.set_option('gui/panels', param)

    def on_gui_loaded(self):

        last_selected_panel = settings.get_option(
            'gui/last_selected_panel', 'collection'
        )

        order = settings.get_option(
            'gui/panels',
            {
                'collection': (True, 0),
                'radio': (True, 1),
                'playlists': (True, 2),
                'files': (True, 3),
            },
        )

        selected_panel = None

        for name, (shown, pos) in order.items():

            panel_data = self.panels.get(name, None)
            if panel_data is None:
                continue

            tab = panel_data.tab
            panel_data.shown = shown
            panel_data.position = pos

            if shown:
                self.reorder_child(tab.page, pos)
            else:
                self.remove_tab(tab)

            if last_selected_panel == name:
                selected_panel = tab.page

        self.loading_panels = False

        # can't determine selected panel when reordering, find it now

        if selected_panel is not None:
            panel_num = self.page_num(selected_panel)
            self.set_current_page(panel_num)


def _register_builtin_panels(exaile, window):

    from xlgui.panel import collection, radio, playlists, files

    logger.info("Loading panels...")

    providers.register(
        'main-panel',
        collection.CollectionPanel(
            window, exaile.collection, 'collection', _show_collection_empty_message=True
        ),
    )

    providers.register(
        'main-panel',
        radio.RadioPanel(
            window, exaile.collection, exaile.radio, exaile.stations, 'radio'
        ),
    )

    providers.register(
        'main-panel',
        playlists.PlaylistsPanel(
            window,
            exaile.playlists,
            exaile.smart_playlists,
            exaile.collection,
            'playlists',
        ),
    )

    providers.register(
        'main-panel', files.FilesPanel(window, exaile.collection, 'files')
    )

    providers.register('main-panel', lyrics.LyricsPanel(window, 'lyrics'))



# Copyright (C) 2009-2010 Erin Drummond
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


class TreeItem:
    def __init__(self, id, name, has_been_expanded=False):
        self._id = id
        self._name = name
        self._type = type
        self._has_been_expanded = has_been_expanded

    def __repr__(self):
        return self._name

    # Getter and Setter crap
    def get_id(self):
        return self._id

    def get_name(self):
        return self._name

    def get_has_been_expanded(self):
        return self._has_been_expanded

    def set_has_been_expanded(self, value):
        self._has_been_expanded = value

    # this is the location in the TreeStore so that child objects can be added to this object (row_pointer should be a TreeIter object)
    def get_row_pointer(self):
        return self._row_pointer

    def set_row_pointer(self, value):
        self._row_pointer = value

    id = property(get_id)
    name = property(get_name)
    expanded = property(get_has_been_expanded, set_has_been_expanded)
    row_pointer = property(get_row_pointer, set_row_pointer)


class Artist(TreeItem):
    def __init__(self, id, name, has_been_expanded=False):
        TreeItem.__init__(self, id, name, has_been_expanded)
        self._albums = []

    # add an album to this artist
    def add_album(self, album):
        if isinstance(album, Album):
            album.artist_name = self.name
            self._albums.append(album)

    def get_albums(self):
        return self._albums

    albums = property(get_albums)


class Album(TreeItem):
    def __init__(self, id, name, has_been_expanded=False):
        TreeItem.__init__(self, id, name, has_been_expanded)
        self._tracks = []
        self._artist_name = "Default"

    def add_track(self, track):
        if isinstance(track, Track):
            track.artist_name = self.artist_name
            track.album_name = self.name
            self._tracks.append(track)

    def get_tracks(self):
        return self._tracks

    def get_artist_name(self):
        return self._artist_name

    def set_artist_name(self, value):
        self._artist_name = value

    tracks = property(get_tracks)
    artist_name = property(get_artist_name, set_artist_name)


class Track(TreeItem):
    def __init__(self, id, name, url):
        TreeItem.__init__(self, id, name)
        self._url = url
        self._artist_name = "http://jamendo.com"
        self._album = "http://jamendo.com"

    def get_url(self):
        return self._url

    def get_album_name(self):
        return self._album

    def set_album_name(self, value):
        self._album = value

    def get_artist_name(self):
        return self._artist_name

    def set_artist_name(self, value):
        self._artist_name = value

    url = property(get_url)
    # used for adding the track to the playlist initially
    artist_name = property(get_artist_name, set_artist_name)
    album_name = property(get_album_name, set_album_name)



import unittest

from mox3 import mox

from gi.repository import Gio

import xl.collection
import xl.trax.search
import xl.trax.track
import xl.trax.util


def test_is_valid_track_valid(test_track):
    fname = test_track.filename
    assert xl.trax.util.is_valid_track(fname), fname


def test_is_valid_track_invalid():
    assert not xl.trax.util.is_valid_track('/')
    assert not xl.trax.util.is_valid_track('/tmp')
    assert not xl.trax.util.is_valid_track(__file__)
    assert not xl.trax.util.is_valid_track('http:///tmp')


class TestGetTracksFromUri:
    class DummyClass:
        def __init__(self, parent, retval):
            self.parent = parent
            self.retval = retval

        def query_info(self, value):
            assert value == 'standard::type'
            return self

        def get_file_type(self):
            return self.retval

    def setup(self):
        self.mox = mox.Mox()

    def teardown(self):
        self.mox.UnsetStubs()

    def get_anything(self, file_type):
        anything = self.mox.CreateMockAnything()
        anything.get_uri_scheme().AndReturn("file")
        if file_type == 'f':
            file_type = Gio.FileType.REGULAR
        elif file_type == 'd':
            file_type = Gio.FileType.DIRECTORY
        elif file_type == 'n':
            #            anything.query_exists(None).AndReturn(False)
            #            return anything
            pass
        else:
            raise NotImplementedError
        #        anything.query_exists(None).AndReturn(True)
        anything.query_info('standard::type').AndReturn(anything)
        anything.get_file_type().AndReturn(file_type)
        return anything

    @unittest.skip("Test is borken because of moxing out error")
    def test_invalid(self):
        loc = '/tmp/foo'
        self.mox.StubOutWithMock(Gio, 'File')
        f_anything = self.get_anything('n')
        Gio.File(loc).AndReturn(f_anything)
        self.mox.ReplayAll()
        assert xl.trax.util.get_tracks_from_uri(loc) == []
        self.mox.VerifyAll()

    @unittest.skip("Test is borken because of moxing out error")
    def test_single(self):
        loc = '/tmp/foo'
        self.mox.StubOutWithMock(Gio, 'FileInfo')
        f_anything = self.mox.CreateMockAnything()
        Gio.FileInfo().AndReturn(f_anything)
        f_anything.get_file_type().AndReturn(Gio.FileType.REGULAR)
        self.mox.ReplayAll()
        assert xl.trax.util.get_tracks_from_uri(loc) == [xl.trax.track.Track(loc)]
        self.mox.VerifyAll()

    @unittest.skip("Test is borken because of moxing out error")
    def test_directory(self):
        loc = '/tmp/foo'
        retval = ['foo', 'bar', 'baz']
        # Gio call to find type
        self.mox.StubOutWithMock(Gio, 'File')
        d_anything = self.get_anything('d')
        Gio.File(loc).AndReturn(d_anything)

        # scanning
        self.mox.StubOutWithMock(xl.collection.Library, 'rescan')
        xl.collection.Library.rescan()
        self.mox.StubOutWithMock(xl.collection.Collection, 'get_tracks')
        xl.collection.Collection.get_tracks().AndReturn(retval)

        self.mox.ReplayAll()
        xl.trax.util.get_tracks_from_uri(loc)
        self.mox.VerifyAll()


class TestSortTracks:
    def setup(self):
        self.tracks = [
            xl.trax.track.Track(url) for url in ('/tmp/foo', '/tmp/bar', '/tmp/baz')
        ]
        for track, val in zip(self.tracks, 'aab'):
            track.set_tag_raw('artist', val)
        for track, val in zip(self.tracks, '212'):
            track.set_tag_raw('discnumber', val)
        self.fields = ('artist', 'discnumber')
        self.result = [self.tracks[1], self.tracks[0], self.tracks[2]]

    def test_sorted(self):
        assert xl.trax.util.sort_tracks(self.fields, self.tracks) == self.result

    def test_reversed(self):
        assert xl.trax.util.sort_tracks(self.fields, self.tracks, reverse=True) == list(
            reversed(self.result)
        )


class TestSortResultTracks:
    def setup(self):
        tracks = [
            xl.trax.track.Track(url) for url in ('/tmp/foo', '/tmp/bar', '/tmp/baz')
        ]
        for track, val in zip(tracks, 'aab'):
            track.set_tag_raw('artist', val)
        for track, val in zip(tracks, '212'):
            track.set_tag_raw('discnumber', val)
        self.tracks = [xl.trax.search.SearchResultTrack(track) for track in tracks]
        self.fields = ('artist', 'discnumber')
        self.result = [self.tracks[1], self.tracks[0], self.tracks[2]]

    def test_sorted(self):
        assert xl.trax.util.sort_result_tracks(self.fields, self.tracks) == self.result

    def test_reversed(self):
        assert xl.trax.util.sort_result_tracks(self.fields, self.tracks, True) == list(
            reversed(self.result)
        )



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Arunas Radzvilavicius, arunas.rv@gmail.com
#    TODO: download books?
#    Drag and drop for books with yet unknown chapters works with
#    main playlist only (see get_chapters_and_drop and related functions).
#    Otherwise, (without threaded functions) UI is non responsive, while
#    fetching that info...

from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import Gtk

import os

from . import librivoxsearch as LS
from . import about_window as AW
from xl import common, event, providers, settings, trax
from xlgui import guiutil, icons, main
from xlgui.widgets.common import DragTreeView
from xlgui.widgets.notebook import NotebookPage


def enable(exaile):
    if exaile.loading:
        event.add_callback(_enable, 'exaile_loaded')
    else:
        _enable(None, exaile, None)


def _enable(o1, exaile, o2):
    global LVPANEL
    LVPANEL = LVPanel(exaile)

    providers.register('main-panel', LVPANEL)


def disable(exaile):
    global LVPANEL
    if LVPANEL.aboutwindow is not None:
        LVPANEL.aboutwindow.win.destroy()

    providers.unregister('main-panel', LVPANEL)


class LVPanel:
    def on_search(self, widget):
        self.run_search(widget)

    @common.threaded
    def run_search(self, widget):
        (c_id, msg_id) = self.statusbar.set_status('Searching...')
        self.keyword = self.entry.get_text()
        self.books = LS.find_books(self.keyword, self._user_agent)
        self.generate_treestore(self.books)
        self.statusbar.unset_status(c_id, msg_id)

    def on_row_exp(self, treeview, iter, path):
        row = path[0]
        if not self.books[row].is_loading and not self.books[row].loaded:
            self.get_chapters(row)

    @common.threaded
    def get_chapters(self, row):
        self.get_all(row)
        return

    @common.threaded
    def get_chapters_and_add(self, row):
        self.get_all(row)
        self.add_to_playlist(self.books[row].chapters)
        return

    @common.idle_add()
    def done_getting_chapters(self, row):
        # adds chapters to treeview and removes "Loading..." message
        l_iter = self.treestore.get_iter((row, 0))
        iter = self.treestore.get_iter((row,))
        for chapter in self.books[row].chapters:
            self.rowlvl2 = self.treestore.append(iter, [chapter[0], self.chapter_icon])
        if len(self.books[row].chapters) > 0:
            self.treestore.remove(l_iter)

    def __init__(self, exaile):

        self.name = 'librivox'  # needed for panel provider
        self._panel = None

        self._user_agent = exaile.get_user_agent_string('librivox')

        self.librivoxdir = os.path.dirname(__file__)
        self.abicon = GdkPixbuf.Pixbuf.new_from_file(self.librivoxdir + '/ebook.png')
        self.clock_icon = GdkPixbuf.Pixbuf.new_from_file(
            self.librivoxdir + '/clock.png'
        )
        self.chapter_icon = icons.MANAGER.pixbuf_from_icon_name(
            'audio-x-generic', Gtk.IconSize.SMALL_TOOLBAR
        )
        self.gui_init(exaile)
        self.connect_events()
        self.getting_info = False

    def gui_init(self, exaile):
        self.vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.vbox.set_border_width(3)
        self.search_label = Gtk.Label(label="LibriVox.org")
        self.vbox.pack_start(self.search_label, False, True, 4)
        self.entry = guiutil.SearchEntry()
        self.entry.connect("activate", self.run_search)

        self.hbox = Gtk.Box()
        self.vbox.pack_start(self.hbox, False, True, 0)
        self.hbox.pack_start(self.entry.entry, True, True, 0)
        self.searchbutton = Gtk.Button()

        self.searchimage = Gtk.Image.new_from_icon_name('edit-find', Gtk.IconSize.MENU)
        self.searchbutton.set_image(self.searchimage)
        self.searchbutton.connect("pressed", self.run_search)
        self.hbox.pack_start(self.searchbutton, False, True, 0)

        self.statusbar = Status()
        self.vbox.pack_start(self.statusbar.bar, False, True, 0)

        self.scrlw = Gtk.ScrolledWindow()
        self.scrlw.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self.scrlw.set_shadow_type(Gtk.ShadowType.IN)
        self.vbox.pack_start(self.scrlw, True, True, 0)

        self.treestore = Gtk.TreeStore(str, GdkPixbuf.Pixbuf)
        self.treeview = Gtk.TreeView.new_with_model(self.treestore)
        self.treeview.set_headers_visible(False)
        self.column = Gtk.TreeViewColumn(None)
        self.cell = Gtk.CellRendererText()
        if settings.get_option('gui/ellipsize_text_in_panels', False):
            from gi.repository import Pango

            self.cell.set_property('ellipsize-set', True)
            self.cell.set_property('ellipsize', Pango.EllipsizeMode.END)
        self.cellpb = Gtk.CellRendererPixbuf()
        self.column.pack_start(self.cellpb, False)
        self.column.pack_start(self.cell, True)
        self.column.set_attributes(self.cell, text=0)
        self.column.set_attributes(self.cellpb, pixbuf=1)
        self.treeview.append_column(self.column)
        self.treeview.connect("row-expanded", self.on_row_exp)

        self.treeview.drag_source_set(
            Gdk.ModifierType.BUTTON1_MASK,
            [Gtk.TargetEntry.new('text/uri-list', 0, 0)],
            Gdk.DragAction.COPY,
        )
        self.treeview.connect("drag-data-get", self.drag_data_get)
        self.scrlw.add(self.treeview)

        self.title = 'LibriVox'
        self.vbox.show_all()

        self.popup_menu = Gtk.Menu()
        self.add_to_pl = Gtk.ImageMenuItem.new_with_label("Append to Current")
        self.add_to_pl.set_image(
            Gtk.Image.new_from_icon_name('list-add', Gtk.IconSize.MENU)
        )
        self.about_book = Gtk.ImageMenuItem.new_with_label("About the Book")
        self.about_book.set_image(
            Gtk.Image.new_from_icon_name('help-about', Gtk.IconSize.MENU)
        )
        self.popup_menu.add(self.add_to_pl)
        self.popup_menu.add(self.about_book)
        self.popup_menu.show_all()

        self.aboutwindow = None

    def connect_events(self):
        self.treeview.connect("row-activated", self.on_append_to_playlist)
        self.treeview.connect("button-press-event", self.menu_popup)
        self.add_to_pl.connect("activate", self.on_append_to_playlist)
        self.about_book.connect("button-press-event", self.on_about_book)
        self.treeview.connect("drag_begin", self.on_drag_begin)

    def generate_tracks(self, chapters):
        tracks = []
        for chapter in chapters:
            chapter_track = trax.Track(chapter[1])
            chapter_track.set_tags(
                artist='Librivox.org', title=chapter[0], album='Audiobook'
            )
            tracks.append(chapter_track)
        return tracks

    def on_append_to_playlist(self, widget, path=None, column=None):
        if not path:
            path = self.treeview.get_cursor()[0]
        row = path[0]
        if self.books[row].is_loading:
            return
        if len(path) == 1:  # selected item is book
            if self.books[row].loaded:  # book info already loaded
                self.add_to_playlist(self.books[row].chapters)
            else:
                self.get_chapters_and_add(row)

        if len(path) > 1:  # selected item is chapter
            chapter = self.books[path[0]].chapters[path[1]]
            self.add_to_playlist([chapter])

    def add_to_playlist(self, chapters):
        current_playlist = main.get_selected_playlist()
        if not current_playlist:
            return

        tracks = self.generate_tracks(chapters)
        current_playlist.playlist.extend(tracks)

    def menu_popup(self, treeview, event):
        if event.triggers_context_menu():
            x = int(event.x)
            y = int(event.y)
            time = event.time
            pthinfo = self.treeview.get_path_at_pos(x, y)
            if pthinfo is not None:
                path, col, cellx, celly = pthinfo
                self.treeview.grab_focus()
                self.treeview.set_cursor(path, col, 0)
                self.popup_menu.popup(None, None, None, None, event.button, time)
            return 1

    @common.idle_add()
    def generate_treestore(self, books):
        self.treestore.clear()
        for book in books:
            self.rowlvl1 = self.treestore.append(None, [book.title, self.abicon])
            self.rowlvl2 = self.treestore.append(
                self.rowlvl1, ["Loading...", self.clock_icon]
            )

    def drag_data_get(self, treeview, context, selection, info, timestamp):
        path = self.treeview.get_cursor()[0]
        if not path:
            return
        if len(path) == 1:  # book selected
            book = self.books[path[0]]
            uris = []
            if book.loaded:  # chapters loaded
                for chapter in book.chapters:
                    chapter_track = self.generate_tracks([chapter])
                    DragTreeView.dragged_data[chapter[1]] = chapter_track[0]
                    uris.append(chapter[1])
                selection.set_uris(uris)
                del uris

            else:
                if book.is_loading:
                    return

                current_playlist = main.get_selected_playlist()
                if not current_playlist:
                    return

                current_playlist_tv = current_playlist.list

                (x, y) = current_playlist_tv.get_pointer()
                rect = current_playlist_tv.get_allocation()
                if x < 0 or x > rect.width or y < 0 or y > rect.height:
                    # dropping not in main playlist
                    return
                drop_info = current_playlist_tv.get_dest_row_at_pos(x, y)

                if drop_info:
                    PLpath, position = drop_info
                    if position == Gtk.TreeViewDropPosition.AFTER:
                        after = True
                    else:
                        after = False

                    self.get_chapters_and_drop(book, current_playlist, PLpath, after)

                else:
                    self.get_chapters_and_add(path[0])

        elif len(path) == 2:
            if self.books[path[0]].is_loading:
                return
            chapter = self.books[path[0]].chapters[path[1]]
            chapter_track = self.generate_tracks([chapter])
            DragTreeView.dragged_data[chapter[1]] = chapter_track[0]
            uri = chapter[1]
            selection.set('text/uri-list', 0, uri)

    @common.threaded
    def get_chapters_and_drop(self, book, current_playlist, PLpath, after):
        path = self.treeview.get_cursor()[0]
        row = path[0]
        self.get_all(row)
        self.drop_after_getting(book, current_playlist, PLpath, after)

    @common.idle_add()
    def drop_after_getting(self, book, current_playlist, PLpath, after):
        # simulates drag_data_received() function of xlgui/playlist.py
        if current_playlist.playlist.ordered_tracks:
            curtrack = current_playlist.playlist.get_current()
        else:
            curtrack = None
        iter = current_playlist.model.get_iter(PLpath)
        tracks = self.generate_tracks(book.chapters)
        for track in tracks:
            ar = current_playlist._get_ar(track)
            if not after:
                after = True
                iter = current_playlist.model.insert_before(iter, ar)
            else:
                iter = current_playlist.model.insert_after(iter, ar)
        # re add to ordered playlist
        current_playlist.playlist.ordered_tracks = []
        for row in current_playlist.model:
            current_playlist.playlist.ordered_tracks.append(row[0])
        current_playlist.main.update_track_counts()
        if curtrack is not None:
            index = current_playlist.playlist.index(curtrack)
            current_playlist.playlist.set_current_pos(index)

    def on_drag_begin(self, widget, context):
        self.treeview.drag_source_set_icon_pixbuf(self.abicon)

    @common.threaded
    def on_about_book(self, widget, event):
        path = self.treeview.get_cursor()[0]
        row = path[0]
        if self.books[row].is_loading:
            return
        if not self.books[row].loaded:
            self.get_all(row)
        self.done_getting_info(row)

    @common.idle_add()
    def done_getting_info(self, row):
        if self.aboutwindow is None:
            self.aboutwindow = AW.AboutWindow()
        self.aboutwindow.set_text(self.books[row])
        if not self.aboutwindow.showing:
            self.aboutwindow.win.show()
            self.aboutwindow.showing = True
        elif self.aboutwindow.showing:
            self.aboutwindow.win.present()

    def get_all(self, row):
        (c_id, msg_id) = self.statusbar.set_status('Loading...')
        self.books[row].is_loading = True
        self.books[row].get_all()
        self.done_getting_chapters(row)
        self.books[row].is_loading = False
        self.statusbar.unset_status(c_id, msg_id)

    def get_panel(self):
        if self._panel is None:
            self._panel = NotebookPage(self.vbox, self.title, 'panel-tab-context')
        return self._panel


class Status:
    '''Status bar'''

    def __init__(self):
        self.bar = Gtk.Statusbar()
        self.bar.show_all()

    def set_status(self, status):
        context_id = 1
        msg_id = self.bar.push(context_id, status)
        return (context_id, msg_id)

    def unset_status(self, context_id, msg_id):
        self.bar.remove(context_id, msg_id)



# Copyright (C) 2008-2010 Adam Olsen
# Copyright (C) 2014-2015 Dustin Spicuzza
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


import time

from gi.repository import GLib

from xl import common
from xl import event
from xl import settings

import logging

logger = logging.getLogger(__name__)


class ExailePlayer:
    """
    This is the player object that everything in Exaile interacts with
    to control audio playback. The player object controls a playback
    engine, which actually controls audio playback. Nothing in this
    object should be specific to a particular engine. Examples of engines
    could be GStreamer, Xine, etc. Currently only the GStreamer engine
    is actually implemented.

    All public functions are assumed to be called from the Glib main
    thread, or bad things will happen. This includes most engine functions,
    with one or two noted exceptions.
    """

    def __init__(self, name, disable_autoswitch=False):
        self.queue = None
        self._name = name

        self._playtime_stamp = None

        self._delay_id = None
        self._stop_id = None
        self._engine = None

        self._auto_advance_delay = 0
        self._auto_advance = True
        self._gapless_enabled = True
        self.__volume = 1.0

        options = {
            '%s/auto_advance_delay' % name: '_auto_advance_delay',
            '%s/auto_advance' % name: '_auto_advance',
            '%s/gapless_playback' % name: '_gapless_enabled',
            '%s/volume' % name: '_volume',
        }

        self._settings_unsub = common.subscribe_for_settings(name, options, self)

        self._setup_engine(disable_autoswitch)

        event.add_callback(self._on_track_end, 'playback_track_end', self)
        event.add_callback(self._on_track_tags_changed, 'track_tags_changed')

    def _setup_engine(self, disable_autoswitch):

        if self._engine is not None:
            self._engine.destroy()

        # allows building docs
        engine_name = settings.get_option("player/engine", "normal")

        if engine_name == 'rtfd_hack':
            return None

        # TODO: support other engines
        from .gst.engine import ExaileGstEngine

        self._engine = ExaileGstEngine(self._name, self, disable_autoswitch)
        self._engine.initialize()

    @property
    def _volume(self):
        return self.__volume

    @_volume.setter
    def _volume(self, value):
        self.__volume = value
        if self._engine is not None:
            self.engine_load_volume()

    def _on_track_end(self, name, obj, track):
        if not track:
            return
        try:
            i = int(track.get_tag_raw('__playcount'))
        except Exception:
            i = 0
        track.set_tags(__playcount=i + 1, __last_played=time.time())

    @common.idle_add()
    def _on_track_tags_changed(self, eventtype, track, tags):
        if '__stopoffset' in tags:
            self._engine.on_track_stopoffset_changed(track)

    def destroy(self):
        """
        Destroys the engine and other resources for this player object.
        Unless you own the object, you probably should never call this.
        """
        if self._settings_unsub is not None:
            self._settings_unsub()
            self._settings_unsub = None

        if self._engine is not None:
            self._engine.destroy()
            self._engine = None

    def get_volume(self):
        """
        Gets the current user volume

        :returns: the volume percentage
        :type: int
        """
        return self._volume * 100

    def set_volume(self, volume):
        """
        Sets the current user volume

        :param volume: the volume percentage
        :type volume: int
        """
        volume = common.clamp(volume, 0, 100)
        settings.set_option("%s/volume" % self._name, volume / 100)

    def modify_volume(self, diff):
        """
        Changes the current user volume

        :param diff: the volume difference (pos or neg) percentage units
        :type volume: int
        """
        v = self.get_volume()
        self.set_volume(v + diff)

    @property
    def current(self):
        return self._engine.get_current_track()

    def play(self, track, start_at=None, paused=False):
        """
        Starts the playback with the provided track
        or stops the playback it immediately if none

        :param track: the track to play or None
        :type track: :class:`xl.trax.Track`
        :param start_at: The offset to start playback at, in seconds
        :param paused: If True, start the track in 'paused' mode

        .. note:: The following :doc:`events </xl/event>` will be emitted by this method:

            * `playback_player_start`: indicates the start of playback overall
            * `playback_track_start`: indicates playback start of a track
        """
        if track is None:
            self.stop()
        else:
            if self.is_stopped():
                event.log_event('playback_player_start', self, track)

            play_args = self._get_play_params(track, start_at, paused, False)
            self._engine.play(*play_args)
            if play_args[2]:
                event.log_event('playback_player_pause', self, track)
                event.log_event("playback_toggle_pause", self, track)

    def stop(self):
        """
        Stops the playback

        .. note:: The following :doc:`events </xl/event>` will be emitted by this method:

            * `playback_player_end`: indicates the end of playback overall
            * `playback_track_end`: indicates playback end of a track
        """
        state = self.get_state()

        if state == 'playing' or state == 'paused':

            self._engine.stop()
            return True
        else:
            logger.debug("Stop ignored when state == %s", state)

        return False

    def pause(self):
        """
        Pauses the playback if playing, does not toggle it

        :returns: True if paused, False otherwise

        .. note:: The following :doc:`events </xl/event>` will be emitted by this method:

            * `playback_player_pause`: indicates that the playback has been paused
            * `playback_toggle_pause`: indicates that the playback has been paused or resumed
        """
        self._cancel_delayed_start()
        if self.is_playing():

            current = self.current
            self._update_playtime(current)
            self._engine.pause()
            self._reset_playtime_stamp()

            event.log_event('playback_player_pause', self, current)
            event.log_event("playback_toggle_pause", self, current)
            return True

        return False

    def unpause(self):
        """
        Resumes the playback if it is paused, does not toggle it

        :returns: True if paused, False otherwise

        .. note:: The following :doc:`events </xl/event>` will be emitted by this method:

            * `playback_player_resume`: indicates that the playback has been resumed
            * `playback_toggle_pause`: indicates that the playback has been paused or resumed
        """
        self._cancel_delayed_start()
        if self.is_paused():

            self._reset_playtime_stamp()
            self._engine.unpause()

            current = self.current
            event.log_event('playback_player_resume', self, current)
            event.log_event("playback_toggle_pause", self, current)
            return True
        return False

    def toggle_pause(self):
        """
        Toggles between playing and paused state. Only valid when playback
        is not stopped.

        :returns: True if toggled, false otherwise

        .. note:: The following :doc:`events </xl/event>` will be emitted by this method:

            * `playback_toggle_pause`: indicates that the playback has been paused or resumed
        """
        state = self.get_state()
        if state == 'paused':
            return self.unpause()
        elif state == 'playing':
            return self.pause()
        else:
            return False

    def seek(self, value):
        """
        Seek to a position in the currently playing stream

        :param value: the position in seconds
        :type value: int
        """
        if self._engine.seek(value):
            event.log_event('playback_seeked', self, value)

    def get_position(self):
        """
        Gets the current playback position of the playing track

        :returns: the playback position in nanoseconds
        :rtype: int
        """
        return self._engine.get_position()

    def get_time(self):
        """
        Gets the current playback time

        :returns: the playback time in seconds
        :rtype: float
        """
        return self.get_position() / 1e9

    def get_progress(self) -> float:
        """
        Gets the current playback progress

        :returns: the playback progress as [0..1]
        """
        try:
            progress = self.get_time() / self.current.get_tag_raw("__length")
        except TypeError:  # track doesn't have duration info
            progress = 0
        except AttributeError:  # no current track
            progress = 0
        else:
            if progress < 0:
                progress = 0
            elif progress > 1:
                progress = 1
        return progress

    def set_progress(self, progress):
        """
        Seeks to the progress position

        :param progress: value ranged at [0..1]
        :type progress: float
        """
        seek_position = 0

        try:
            length = self.current.get_tag_raw('__length')
            seek_position = length * progress
        except (TypeError, AttributeError):
            pass

        self.seek(seek_position)

    def modify_time(self, diff):
        """
        Modifies the current position backwards or forwards.

        If position ends up after the end or before the start of the track
        it is truncated to lay inside the track.

        :param diff: value in seconds
        :type diff: int
        """
        try:
            length = self.current.get_tag_raw('__length')
        except (TypeError, AttributeError):
            return

        if length is None:
            return

        pos = self.get_time()
        seek_pos = pos + diff

        # Make sure we don't seek outside the current track. Subtract a little
        # from length, player seems to hang sometimes if we seek to the end of
        # a track.
        seek_pos = max(0, min(seek_pos, length - 2))

        self.seek(seek_pos)

    def get_state(self):
        """
        Gets the player state

        :returns: one of *playing*, *paused* or *stopped*
        :rtype: string
        """
        return self._engine.get_state()

    def is_playing(self):
        """
        Convenience method to find out if the player is currently playing

        :returns: whether the player is currently playing
        :rtype: bool
        """
        return self._engine.get_state() == 'playing'

    def is_paused(self):
        """
        Convenience method to find out if the player is currently paused

        :returns: whether the player is currently paused
        :rtype: bool
        """
        return self._engine.get_state() == 'paused'

    def is_stopped(self):
        """
        Convenience method to find out if the player is currently stopped

        :returns: whether the player is currently stopped
        :rtype: bool
        """
        return self._engine.get_state() == 'stopped'

    #
    # Engine-only API
    #

    def engine_load_volume(self):
        """
        Load volume from settings, this function will call set_volume
        on the engine

        .. note:: Only to be called from engine
        """
        self._engine.set_volume(self._volume)

    def engine_notify_user_volume_change(self, vol):
        """
        Engine calls this when something inside the engine changes
        the user volume.

        .. note:: Only to be called from engine
        """
        settings.set_option('%s/volume' % self._name, vol)

    def engine_notify_track_start(self, track):
        """
        Called when a track has just entered the playing state

        :param track: Track that is being played now

        .. note:: Only to be called from engine
        """

        self._reset_playtime_stamp()
        event.log_event('playback_track_start', self, track)

    def engine_notify_track_end(self, track, done):
        """
        Called when a track has been stopped. Either:

        - stop() was called
        - play() was called and the prior track was stopped

        :param track: Must be the track that was just playing, and must
                      never be None

        :param done:  If True, no further tracks will be played

        .. note:: Only to be called from engine
        """

        self._update_playtime(track)
        event.log_event('playback_track_end', self, track)

        if done:
            self._cancel_delayed_start()
            event.log_event('playback_player_end', self, track)

    @common.idle_add()
    def engine_notify_error(self, msg):
        """
        Notification that some kind of error has occurred. If the error
        is not recoverable, the engine is expected to stop playback and
        reset itself to a state where playback can begin again.

        .. note:: Only to be called from engine
        """
        event.log_event('playback_error', self, msg)

    def engine_autoadvance_get_next_track(self, gapless=False):
        """
        Engine calls this when it wants to see what the next track
        to play is. The track may or may not actually get played.
        If the track gets played, then the engine must call
        engine_autoadvance_notify_next, engine_notify_playback_stop,
        and engine_notify_playback_start

        May be called on another thread.

        :param gapless: Set to True if the autoadvance is part of a gapless
                        playback attempt

        :returns: The next track, or None if the engine should not try
                  to play another track immediately

        .. note:: Only to be called from engine
        """

        if not self._auto_advance:
            return

        if gapless:
            if self._auto_advance_delay != 0 or not self._gapless_enabled:
                return

        return self.queue.get_next()

    def engine_autoadvance_notify_next(self, track):
        """
        Engine calls this when it has started playing the next track as
        part of an auto advance action.

        :param track: The track that will be played

        :returns: (track, start_at, paused) parameters that can be passed
                  to the 'play' function of the engine.

        .. note:: Only to be called from engine
        """

        self.queue.next(autoplay=False)

        return self._get_play_params(track, None, False, True)

    #
    # Playtime related stuffs
    #

    def _get_play_params(self, track, start_at, paused, autoadvance):
        if start_at is None or start_at <= 0:
            start_at = None
            start_offset = track.get_tag_raw('__startoffset') or 0
            if start_offset > 0:
                start_at = start_offset

        # Once playback has started, if there's a delay, pause the stream
        # for delay number of seconds
        self._cancel_delayed_start()

        if not paused and autoadvance and self._auto_advance_delay > 0:
            delay = int(self._auto_advance_delay)
            logger.debug("Delaying start for %sms", delay)
            self._delay_id = GLib.timeout_add(delay, self._delayed_start)
            paused = True

        return track, start_at, paused

    def _delayed_start(self):
        logger.debug("Resuming playback after delayed start")
        self.unpause()

    def _update_playtime(self, track):
        """
        updates the total playtime for the currently playing track

        .. should be called whenever a pause/stop event occurs
        """
        if track and self._playtime_stamp:
            last = track.get_tag_raw('__playtime')
            if isinstance(last, str):
                try:
                    last = int(last)
                except Exception:
                    last = 0
            elif not isinstance(last, int):
                last = 0
            track.set_tag_raw(
                '__playtime', last + int(time.time() - self._playtime_stamp)
            )
            self._playtime_stamp = None

    def _reset_playtime_stamp(self):
        self._playtime_stamp = int(time.time())

    #
    # Delayed start stuff
    #

    def _cancel_delayed_start(self):
        if self._delay_id is not None:
            GLib.source_remove(self._delay_id)
            self._delay_id = None



# Copyright (C) 2006-2010  Johannes Sasongko <sasongko@gmail.com>
# Copyright (C) 2010  Mathias Brodala <info@noctus.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
from xlgui.preferences import widgets
from xl.nls import gettext as _

name = _('Desktop Cover')
basedir = os.path.dirname(os.path.realpath(__file__))
ui = os.path.join(basedir, "desktopcover_preferences.ui")
icon = 'image-x-generic'


class AnchorPreference(widgets.ComboPreference):
    default = 'topleft'
    name = 'plugin/desktopcover/anchor'


class XPreference(widgets.SpinPreference):
    default = 0
    name = 'plugin/desktopcover/x'


class YPreference(widgets.SpinPreference):
    default = 0
    name = 'plugin/desktopcover/y'


class OverrideSizePreference(widgets.CheckPreference):
    default = False
    name = 'plugin/desktopcover/override_size'


class SizePreference(widgets.SpinPreference, widgets.CheckConditional):
    default = 200
    name = 'plugin/desktopcover/size'
    condition_preference_name = 'plugin/desktopcover/override_size'

    def __init__(self, preferences, widget):
        widgets.SpinPreference.__init__(self, preferences, widget)
        widgets.CheckConditional.__init__(self)


class FadingPreference(widgets.CheckPreference):
    default = False
    name = 'plugin/desktopcover/fading'


class FadingDurationPreference(widgets.SpinPreference, widgets.CheckConditional):
    default = 50
    name = 'plugin/desktopcover/fading_duration'
    condition_preference_name = 'plugin/desktopcover/fading'

    def __init__(self, preferences, widget):
        widgets.SpinPreference.__init__(self, preferences, widget)
        widgets.CheckConditional.__init__(self)


# vi: et sts=4 sw=4 tw=80
